<?php

namespace Kompetenztestde\TestcenterConnector\Api\Schemas;

use DateTime;
use Exception;
use Kompetenztestde\TestcenterConnector\Api\FieldValidationConstraint;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;

class SysCheckReport
{
    use ValidationTrait;

    /**
     * @var DateTime
     */
    private $date;

    /**
     * @var string
     */
    private $checkId;

    /**
     * @var string
     */
    private $checkLabel;

    /**
     * @var string
     */
    private $title;

    /**
     * @var SysCheckItem[]|null
     */
    private $environment;

    /**
     * @var SysCheckItem[]|null
     */
    private $network;

    /**
     * @var SysCheckItem[]|null
     */
    private $questionnaire;

    /**
     * @var SysCheckItem[]|null
     */
    private $unit;

    /**
     * @var SysCheckFileData[]|null
     */
    private $fileData;

    /**
     * @param object $data
     * @throws TestcenterConnectorException
     */
    public function __construct(object $data)
    {
        self::validate($data);

        try {
            $date = new DateTime($data->date);
        } catch (Exception $e) {
            throw new TestcenterConnectorException('Expected "date" to be a valid datetime value');
        }

        $environment = isset($data->environment) ? self::getSysCheckItemArray($data->environment) : null;
        $network = isset($data->network) ? self::getSysCheckItemArray($data->network) : null;
        $questionnaire = isset($data->questionnaire) ? self::getSysCheckItemArray($data->questionnaire) : null;
        $unit = isset($data->unit) ? self::getSysCheckItemArray($data->unit) : null;

        $fileData = [];
        if (isset($data->fileData)) {
            foreach ($data->fileData as $fileDataData) {
                if (!is_object($fileDataData)) {
                    throw new TestcenterConnectorException('Expected fileData item to be an object');
                }

                $fileData[] = new SysCheckFileData($fileDataData);
            }
        }

        $this->date = $date;
        $this->checkId = $data->checkId;
        $this->checkLabel = $data->checkLabel;
        $this->title = $data->title;
        $this->environment = $environment;
        $this->network = $network;
        $this->questionnaire = $questionnaire;
        $this->unit = $unit;
        $this->fileData = $fileData;
    }

    /**
     * @return DateTime
     */
    public function getDate(): DateTime
    {
        return $this->date;
    }

    /**
     * @return string
     */
    public function getCheckId(): string
    {
        return $this->checkId;
    }

    /**
     * @return string
     */
    public function getCheckLabel(): string
    {
        return $this->checkLabel;
    }

    /**
     * @return string
     */
    public function getTitle(): string
    {
        return $this->title;
    }

    /**
     * @return SysCheckItem[]
     */
    public function getEnvironment(): array
    {
        return $this->environment;
    }

    /**
     * @return SysCheckItem[]
     */
    public function getNetwork(): array
    {
        return $this->network;
    }

    /**
     * @return SysCheckItem[]
     */
    public function getQuestionnaire(): array
    {
        return $this->questionnaire;
    }

    /**
     * @return SysCheckItem[]
     */
    public function getUnit(): array
    {
        return $this->unit;
    }

    /**
     * @return SysCheckFileData[]
     */
    public function getFileData(): array
    {
        return $this->fileData;
    }

    protected static function getConstraints(): array
    {
        return [
            new FieldValidationConstraint('date', 'regex', true, '/\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}/'),
            new FieldValidationConstraint('checkId', 'string', true),
            new FieldValidationConstraint('checkLabel', 'string', true),
            new FieldValidationConstraint('title', 'string', true),
            new FieldValidationConstraint('environment', 'array', false),
            new FieldValidationConstraint('network', 'array', false),
            new FieldValidationConstraint('questionnaire', 'array', false),
            new FieldValidationConstraint('unit', 'array', false),
            new FieldValidationConstraint('fileData', 'array', false)
        ];
    }

    /**
     * @param array $data
     * @return SysCheckItem[]
     * @throws TestcenterConnectorException
     */
    private static function getSysCheckItemArray(array $data): array
    {
        $items = [];
        foreach ($data as $itemData) {
            if (!is_object($itemData)) {
                throw new TestcenterConnectorException('Expected sys check item to be an object');
            }
            $items[] = new SysCheckItem($itemData);
        }

        return $items;
    }
}
