<?php

namespace Kompetenztestde\TestcenterConnector\Api;

use GuzzleHttp;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Promise\PromiseInterface;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileDeletionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileGetResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\Result;
use Kompetenztestde\TestcenterConnector\Api\Schemas\StartSessionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileUploadResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileValidationResult;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UnitResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\LogEntry;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SysCheckReport;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SystemConfigResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\User;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserRole;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserWorkspace;
use Kompetenztestde\TestcenterConnector\Api\Schemas\Workspace;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SessionResponse;
use Kompetenztestde\TestcenterConnector\Exceptions\ApiException;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;
use Psr\Http\Message\ResponseInterface;

class TestcenterApiClient
{
    /**
     * @var string
     */
    private $baseUrl;

    /**
     * @var GuzzleHttp\Client
     */
    private $httpClient;

    /**
     * @param string $baseUrl
     * @param Client|null $httpClient
     */
    public function __construct(string $baseUrl, GuzzleHttp\Client $httpClient = null)
    {
        $this->baseUrl = $baseUrl;

        if ($httpClient === null) {
            $this->httpClient = new GuzzleHttp\Client();
        } else {
            $this->httpClient = $httpClient;
        }
    }

    /**
     * Starts a Session as Admin by Username and password
     *
     * @param string $username
     * @param string $password
     * @return StartSessionResponse
     * @throws TestcenterConnectorException
     * @throws ApiException
     */
    public function startAdminSession(string $username, string $password): StartSessionResponse
    {
        return $this->startAdminSessionAsync($username, $password)->wait();
    }

    /**
     * Asynchronously start a session as admin by username and password
     *
     * @param string $username
     * @param string $password
     * @return PromiseInterface resolves to {@see StartSessionResponse} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function startAdminSessionAsync(string $username, string $password): PromiseInterface
    {
        $body = [
            'name' => $username,
            'password' => $password
        ];

        $promise = $this->httpClient->putAsync($this->baseUrl . '/session/admin', [
            'json' => $body
        ]);

        return $promise->then(function (ResponseInterface $response) {
            // TODO: Testcenter API documentation says status code 202 but in source it says 204
            if ($response->getStatusCode() === 204) {
                throw new TestcenterConnectorException('User exists but has no admin privileges.');
            }

            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return new StartSessionResponse($data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Uploads a Resource, Unit, Booklet, SysCheck or Testtakers file
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string $filename
     * @param string $content
     * @return FileUploadResponse
     * @throws TestcenterConnectorException
     * @throws ApiException
     */
    public function uploadFile(
        string $authToken,
        int $workspaceId,
        string $filename,
        string $content
    ): FileUploadResponse {
        return $this->uploadFileAsync($authToken, $workspaceId, $filename, $content)->wait();
    }

    /**
     * Asynchronously uploads a Resource, Unit, Booklet, SysCheck or Testtakers file
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string $filename
     * @param string $content
     * @return PromiseInterface resolves to {@see FileUploadResponse} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function uploadFileAsync(
        string $authToken,
        int $workspaceId,
        string $filename,
        string $content
    ): PromiseInterface {
        $requestOptions = [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'multipart' => [
                [
                    'name' => 'fileforvo',
                    'contents' => $content,
                    'filename' => $filename
                ]
            ]
        ];

        $promise = $this->httpClient->postAsync(
            $this->baseUrl . '/workspace/' . $workspaceId . '/file',
            $requestOptions
        );

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            $fileValidationResults = [];
            foreach ($data as $filename => $fileResponse) {
                $errors = $fileResponse->error ?? [];
                $warnings = $fileResponse->warning ?? [];
                $result = new FileValidationResult($filename, $warnings, $errors);
                $fileValidationResults[] = $result;
            }

            return new FileUploadResponse($fileValidationResults);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get information about files of a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @return FileGetResponse
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getFilesOfWorkspace(string $authToken, int $workspaceId): FileGetResponse
    {
        return $this->getFilesOfWorkspaceAsync($authToken, $workspaceId)->wait();
    }

    /**
     * Asynchronously get information about files of a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @return PromiseInterface
     */
    public function getFilesOfWorkspaceAsync(string $authToken, int $workspaceId): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/workspace/{$workspaceId}/files", [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return new FileGetResponse($data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Delete one or more files in a workspace
     *
     * @param string $authToken admin authentication token with RW privileges in the workspace
     * @param int $workspaceId id of the workspace to delete files from
     * @param array $filenames list of filenames with filetype prefix (e.g. 'Testtakers/')
     * @return FileDeletionResponse
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function deleteFiles(string $authToken, int $workspaceId, array $filenames): FileDeletionResponse
    {
        return $this->deleteFilesAsync($authToken, $workspaceId, $filenames)->wait();
    }

    /**
     * Asynchronously delete one or more files in a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param array $filenames
     * @return PromiseInterface resolves to {@see FileDeletionResponses} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function deleteFilesAsync(
        string $authToken,
        int $workspaceId,
        array $filenames
    ): PromiseInterface {
        $promise = $this->httpClient->deleteAsync("{$this->baseUrl}/workspace/$workspaceId/files", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => [
                'f' => $filenames
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return new FileDeletionResponse($data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Add a new workspace to the testcenter
     *
     * @param string $authToken super admin authentication token
     * @param string $name name of the new workspace
     * @return int|null
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function addAWorkspace(string $authToken, string $name): ?int
    {
        return $this->addAWorkspaceAsync($authToken, $name)->wait();
    }

    /**
     * Asynchronously add a new workspace to the testcenter
     *
     * @param string $authToken
     * @param string $name
     * @return PromiseInterface resolves to workspace id (int) or null or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function addAWorkspaceAsync(string $authToken, string $name): PromiseInterface
    {
        $body = [
            'name' => $name
        ];
        $promise = $this->httpClient->putAsync($this->baseUrl . '/workspace', [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => $body
        ]);

        return $promise->then(function (ResponseInterface $response) {
            // check if workspace id is returned (version 15.1.7+)
            $body = $response->getBody()->getContents();
            $workspaceId = filter_var($body, FILTER_VALIDATE_INT);

            if ($workspaceId !== false) {
                return $workspaceId;
            } else {
                return null;
            }
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get an array of all workspaces in the testcenter
     *
     * @param string $authToken super admin authentication token
     * @return Workspace[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getAListOfWorkspaces(string $authToken): array
    {
        return $this->getAListOfWorkspacesAsync($authToken)->wait();
    }

    /**
     * Asynchronously get an array of all workspaces in the testcenter
     *
     * @param string $authToken super admin authentication token
     * @return PromiseInterface resolves to array of {@see Workspace} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function getAListOfWorkspacesAsync(string $authToken): PromiseInterface
    {
        $promise = $this->httpClient->getAsync($this->baseUrl . '/workspaces', [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            $workspaces = [];

            foreach ($data as $workspaceData) {
                $workspaces[] = new Workspace($workspaceData);
            }

            return $workspaces;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get a list of workspaces a user has access to
     *
     * @param string $authToken
     * @param int $userId
     * @return UserWorkspace[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getAListOfUserWorkspaces(string $authToken, int $userId): array
    {
        return $this->getAListOfUserWorkspacesAsync($authToken, $userId)->wait();
    }

    /**
     * Asynchronously get a list of workspaces a user has access to
     *
     * @param string $authToken
     * @param int $userId
     * @return PromiseInterface resolves to array of {@see UserWorkspace} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function getAListOfUserWorkspacesAsync(string $authToken, int $userId): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/user/$userId/workspaces", [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);
        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            if ($data === null) {
                throw new TestcenterConnectorException('Invalid json response');
            }

            if (!is_array($data)) {
                throw new TestcenterConnectorException('Expected response to be an array');
            }

            $workspaces = [];

            foreach ($data as $workspaceData) {
                if (!is_object($workspaceData)) {
                    throw new TestcenterConnectorException('Expected workspace data to be an object');
                }
                $workspaces[] = new UserWorkspace($workspaceData);
            }

            return $workspaces;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Set roles of one or multiple users in a workspace
     *
     * @param string $authToken super admin authentication token
     * @param int $workspaceId id of the workspace to set the roles in
     * @param UserRole[] $userRoles
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function changeUserRoles(string $authToken, int $workspaceId, array $userRoles): bool
    {
        return $this->changeUserRolesAsync($authToken, $workspaceId, $userRoles)->wait();
    }

    /**
     * Asynchronously set roles of one or multiple users in a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param array $userRoles
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function changeUserRolesAsync(
        string $authToken,
        int $workspaceId,
        array $userRoles
    ): PromiseInterface {
        $body = [
            'u' => $userRoles
        ];

        $promise = $this->httpClient->patchAsync($this->baseUrl . '/workspace/' . $workspaceId . '/users', [
            'json' => $body,
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Returns session data according to an authToken
     *
     * @param string $authToken
     * @return SessionResponse
     * @throws TestcenterConnectorException
     * @throws ApiException
     */
    public function getASession(string $authToken): SessionResponse
    {
        return $this->getASessionAsync($authToken)->wait();
    }

    /**
     * Asynchronously returns session data according to an authToken
     *
     * @param string $authToken
     * @return PromiseInterface resolves to {@see SessionResponse} or rejects with {@see TestcenterConnectorException}
     *                          or {@see ApiException}
     */
    public function getASessionAsync(string $authToken): PromiseInterface
    {
        $promise = $this->httpClient->getAsync($this->baseUrl . '/session', [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);
        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return new SessionResponse($data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get responses from a workspace for a list of test groups
     *
     * @param string $authToken admin auth token
     * @param int $workspaceId id of the workspace to get responses from
     * @param string[] $dataIds ids of the test groups to get responses from
     * @return UnitResponse[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getReportOfItemResponses(string $authToken, int $workspaceId, array $dataIds): array
    {
        return $this->getReportOfItemResponsesAsync($authToken, $workspaceId, $dataIds)->wait();
    }

    /**
     * Asynchronously get responses from a workspace for a list of test groups
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $dataIds
     * @return PromiseInterface resolves to array of {@see UnitResponse} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function getReportOfItemResponsesAsync(
        string $authToken,
        int $workspaceId,
        array $dataIds
    ): PromiseInterface {
        $dataIds = implode(',', $dataIds);
        $promise = $this->httpClient->getAsync($this->baseUrl . '/workspace/' . $workspaceId . '/report/response', [
            'headers' => [
                'AuthToken' => $authToken,
                'Accept' => 'application/json'
            ],
            'query' => [
                'dataIds' => $dataIds
            ]
        ]);
        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            if ($data !== null) {
                $responses = array_map(function ($responsesData) {
                    return new UnitResponse($responsesData);
                }, $data);
            } else {
                $responses = [];
            }

            return $responses;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get logs for login groups in a given workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $dataIds
     * @return LogEntry[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getReportOfLogs(string $authToken, int $workspaceId, array $dataIds): array
    {
        return $this->getReportOfLogsAsync($authToken, $workspaceId, $dataIds)->wait();
    }

    /**
     * Asynchronously get logs for login groups in a given workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $dataIds
     * @return PromiseInterface resolves to array of {@see LogEntry} or rejects with {@see TestcenterConnectorException}
     *                          or {@see ApiException}
     */
    public function getReportOfLogsAsync(
        string $authToken,
        int $workspaceId,
        array $dataIds
    ): PromiseInterface {
        $promise = $this->httpClient->getAsync($this->baseUrl . '/workspace/' . $workspaceId . '/report/log', [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'query' => [
                'dataIds' => implode(',', $dataIds)
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            // testcenter returns an empty body if there are no log entries instead of an empty array
            if ($data === null) {
                $logEntries = [];
            } else {
                $logEntries = array_map(function ($entryData) {
                    return new LogEntry($entryData);
                }, $data);
            }

            return $logEntries;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Add a new user
     *
     * @param string $authToken super admin authentication token
     * @param string $username name of the new user
     * @param string $password password of the new user
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function addAUser(string $authToken, string $username, string $password): bool
    {
        return $this->addAUserAsync($authToken, $username, $password)->wait();
    }

    /**
     * Asynchronously add a new user
     *
     * @param string $authToken
     * @param string $username
     * @param string $password
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function addAUserAsync(
        string $authToken,
        string $username,
        string $password
    ): PromiseInterface {
        $body = [
            'n' => $username,
            'p' => $password
        ];

        $promise = $this->httpClient->putAsync("{$this->baseUrl}/user", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => $body
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get a list of all registered users
     *
     * @param string $authToken
     * @return User[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getAListOfUsers(string $authToken): array
    {
        return $this->getAListOfUsersAsync($authToken)->wait();
    }

    /**
     * Asynchronously get a list of all registered users
     *
     * @param string $authToken
     * @return PromiseInterface resolves to array of {@see User} or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function getAListOfUsersAsync(string $authToken): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/users", [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            if ($data === null) {
                throw new TestcenterConnectorException('Invalid json response');
            }

            if (!is_array($data)) {
                throw new TestcenterConnectorException('Expected array response');
            }

            $users = [];

            foreach ($data as $userData) {
                if (!is_object($userData)) {
                    throw new TestcenterConnectorException('Expected object user data');
                }

                $users[] = new User($userData);
            }

            return $users;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Change the password of a user
     *
     * @param string $authToken
     * @param int $userId
     * @param string $newPassword
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function changeUserPassword(string $authToken, int $userId, string $newPassword): bool
    {
        return $this->changeUserPasswordAsync($authToken, $userId, $newPassword)->wait();
    }

    /**
     * Asynchronously change the password of a user
     *
     * @param string $authToken
     * @param int $userId
     * @param string $newPassword
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function changeUserPasswordAsync(string $authToken, int $userId, string $newPassword): PromiseInterface
    {
        $promise = $this->httpClient->patchAsync("{$this->baseUrl}/user/$userId/password", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => [
                'p' => $newPassword
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Change the super admin status of a user
     *
     * @param string $authToken super admin authentication token
     * @param int $userId id of the user to change the status of
     * @param bool $newStatus true to enable super admin, false to disable
     * @param string $password password of the currently logged in super admin
     * @return bool true on success, false if not
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function changeSuperAdminStatus(string $authToken, int $userId, bool $newStatus, string $password): bool
    {
        return $this->changeSuperAdminStatusAsync($authToken, $userId, $newStatus, $password)->wait();
    }

    /**
     * Asynchronously change the super admin status of a user
     *
     * @param string $authToken
     * @param int $userId
     * @param bool $newStatus
     * @param string $password
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function changeSuperAdminStatusAsync(
        string $authToken,
        int $userId,
        bool $newStatus,
        string $password
    ): PromiseInterface {
        $newStatus = $newStatus ? 'on' : 'off';
        $promise = $this->httpClient->patchAsync("{$this->baseUrl}/user/$userId/super-admin/$newStatus", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => [
                'p' => $password
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Delete one or more users
     *
     * @param string $authToken
     * @param int[] $userIds ids of users to delete
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function deleteSomeUsers(string $authToken, array $userIds): bool
    {
        return $this->deleteSomeUsersAsync($authToken, $userIds)->wait();
    }

    /**
     * Asynchronously delete one or more users
     * @param string $authToken
     * @param array $userIds
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function deleteSomeUsersAsync(string $authToken, array $userIds): PromiseInterface
    {
        $promise = $this->httpClient->deleteAsync("{$this->baseUrl}/users", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => [
                'u' => $userIds
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Delete one or many workspaces
     *
     * @param string $authToken authentication token of a super admin user
     * @param int[] $workspaceIds array of workspace ids to delete
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function deleteSomeWorkspaces(string $authToken, array $workspaceIds): bool
    {
        return $this->deleteSomeWorkspacesAsync($authToken, $workspaceIds)->wait();
    }

    /**
     * Asynchronously delete one or many workspaces
     *
     * @param string $authToken
     * @param array $workspaceIds
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function deleteSomeWorkspacesAsync(string $authToken, array $workspaceIds): PromiseInterface
    {
        $requestBody = [
            'ws' => $workspaceIds
        ];

        $promise = $this->httpClient->deleteAsync("{$this->baseUrl}/workspaces", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => $requestBody
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return ($response->getStatusCode() >= 200 && $response->getStatusCode() < 300);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }


    /**
     * Rename a workspace
     *
     * @param string $authToken authentication token for super admin user
     * @param int $workspaceId id of workspace to rename
     * @param string $name new name
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function renameAWorkspace(string $authToken, int $workspaceId, string $name): bool
    {
        return $this->renameAWorkspaceAsync($authToken, $workspaceId, $name)->wait();
    }

    /**
     * Asynchronously rename a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string $name
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function renameAWorkspaceAsync(string $authToken, int $workspaceId, string $name): PromiseInterface
    {
        $body = [
            'name' => $name
        ];

        $promise = $this->httpClient->patchAsync("{$this->baseUrl}/workspace/$workspaceId", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => $body
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get statistical results from a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @return Result[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getResults(string $authToken, int $workspaceId): array
    {
        return $this->getResultsAsync($authToken, $workspaceId)->wait();
    }

    /**
     * Asynchronously get statistical results from a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @return PromiseInterface resolves to array of {@see Result} or rejects with {@see TestcenterConnectorException}
     *                          or {@see ApiException}
     */
    public function getResultsAsync(string $authToken, int $workspaceId): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/workspace/$workspaceId/results", [
            'headers' => [
                'AuthToken' => $authToken
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            $results = [];
            foreach ($data as $resultData) {
                $results[] = new Result($resultData);
            }

            return $results;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Returns a System Check report
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $dataIds
     * @return SysCheckReport[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getReportOfSystemChecks(string $authToken, int $workspaceId, array $dataIds): array
    {
        return $this->getReportOfSystemChecksAsync($authToken, $workspaceId, $dataIds)->wait();
    }

    /**
     * Asynchronously returns a System Check report
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $dataIds
     * @return PromiseInterface resolves to array of {@see SysCheckReport} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function getReportOfSystemChecksAsync(string $authToken, int $workspaceId, array $dataIds): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/workspace/$workspaceId/report/sys-check", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'query' => [
                'dataIds' => implode(',', $dataIds)
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return array_map(function ($reportData) {
                return new SysCheckReport($reportData);
            }, $data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Returns publicly available parts of system config:
     * version number, customTexts for UI, broadcastingService infos, serverTimestamp
     *
     * @return SystemConfigResponse
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function getSystemConfig(): SystemConfigResponse
    {
        return $this->getSystemConfigAsync()->wait();
    }

    /**
     * Asynchronously returns publicly available parts of system config:
     * version number, customTexts for UI, broadcastingService infos, serverTimestamp
     *
     * @return PromiseInterface resolves to {@see SystemConfigResponse} or rejects with
     *                          {@see TestcenterConnectorException} or {@see ApiException}
     */
    public function getSystemConfigAsync(): PromiseInterface
    {
        $promise = $this->httpClient->getAsync("{$this->baseUrl}/system/config");

        return $promise->then(function (ResponseInterface $response) {
            $body = $response->getBody()->getContents();
            $data = json_decode($body);

            return new SystemConfigResponse($data);
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Delete responses from a workspace
     *
     * @param string $authToken authentication token of admin user with RW permissions
     * @param int $workspaceId id of the workspace to delete data from
     * @param string[] $groups group ids to delete data from
     * @return bool
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    public function deleteData(string $authToken, int $workspaceId, array $groups): bool
    {
        return $this->deleteDataAsync($authToken, $workspaceId, $groups)->wait();
    }

    /**
     * Asynchronously delete responses from a workspace
     *
     * @param string $authToken
     * @param int $workspaceId
     * @param string[] $groups
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException} or
     *                          {@see ApiException}
     */
    public function deleteDataAsync(string $authToken, int $workspaceId, array $groups): PromiseInterface
    {
        $promise = $this->httpClient->deleteAsync("{$this->baseUrl}/workspace/$workspaceId/responses", [
            'headers' => [
                'AuthToken' => $authToken
            ],
            'json' => [
                'groups' => $groups
            ]
        ]);

        return $promise->then(function (ResponseInterface $response) {
            return true;
        }, function (GuzzleException $e) {
            $this->handleException($e);
        });
    }

    /**
     * Get the testcenter api base url
     *
     * @return string
     */
    public function getBaseUrl(): string
    {
        return $this->baseUrl;
    }

    /**
     * Handle guzzle exceptions
     *
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    private function handleException(GuzzleException $e): void
    {
        if ($e instanceof RequestException) {
            throw new ApiException(
                "[{$e->getCode()}] {$e->getMessage()}",
                $e->getCode(),
                $e->getResponse() ? $e->getResponse()->getHeaders() : null,
                $e->getResponse() ? $e->getResponse()->getBody()->getContents() : null
            );
        } else {
            throw new TestcenterConnectorException(
                'Failed to connect to testcenter server (Reason: ' . $e->getMessage() . ')'
            );
        }
    }
}
