<?php

namespace Kompetenztestde\TestcenterConnector;

use Exception;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileDeletionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileUploadResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserRole;
use Kompetenztestde\TestcenterConnector\Models\File;
use Kompetenztestde\TestcenterConnector\Models\TestGroup;
use Kompetenztestde\TestcenterConnector\Models\TestInterface;
use Kompetenztestde\TestcenterConnector\Models\Workspace;
use Kompetenztestde\TestcenterConnector\Xml\XmlGenerator;
use Throwable;
use GuzzleHttp\Client;
use GuzzleHttp\Promise\Utils;
use LogicException;
use BadMethodCallException;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;

// phpcs:disable
/**
 * Wrapper class for {@see TestcenterConnector} to communicate with multiple testcenter instances in parallel
 *
 * Async methods are called in parallel other methods have to get called one after another. When working with multiple
 * testcenter instances the use of async methods is recommended for improved performance.
 *
 * On success the methods return an associative array of the return values of each instance with the instance base url
 * as keys. If any call fails a {@see TestcenterConnectorException} is thrown
 *
 * @method FileUploadResponse[] addTestGroup(int|string $workspace, TestGroup $group, string $bookletId, XmlGenerator $xmlGenerator = null)
 * @method FileUploadResponse[] addTestGroupAsync(int|string $workspace, TestGroup $group, string $bookletId, XmlGenerator $xmlGenerator = null)
 * @method int[] addWorkspace(string $name, ?bool $setRole = false)
 * @method int[] addWorkspaceAsync(string $name, ?bool $setRole = false)
 * @method Workspace[] getWorkspaces()
 * @method Workspace[] getWorkspacesAsync()
 * @method bool[] deleteWorkspace(int|string[] $workspaces)
 * @method bool[] deleteWorkspaceAsync(int|string[] $workspaces)
 * @method FileUploadResponse[] addTest(int|string $workspace, TestInterface $test)
 * @method array getGroupResults(int|string $workspace, string $groupId)
 * @method array getGroupResultsAsync(int|string $workspace, string $groupId)
 * @method bool[] deleteUsers(int|string[] $users)
 * @method bool[] addUser(string $username, string $password)
 * @method int[] addSuperAdmin(string $username, string $password)
 * @method array getUsers()
 * @method bool[] renameWorkspace(int|string $workspace, string $name)
 * @method bool[] deleteData(int|string $workspace, string[] $groups)
 * @method array getSystemCheckReport(int|string $workspace, string[] $dataIds)
 * @method array getResults(int|string $workspace)
 * @method array getResultsAsync(int|string $workspace)
 * @method array getWorkspacesByUser(int|string $user)
 * @method bool[] changeUserRoles(int|string $workspace, UserRole[] $userRoles)
 * @method array getLogs(int|string $workspace, string[] $groupIds)
 * @method array getLogsAsync(int|string $workspace, string[] $groupIds)
 * @method FileDeletionResponse[] deleteFiles(int|string $workspace, File[] $files)
 * @method bool[] changeUserPassword(int|string $user, string $newPassword)
 */
// phpcs:enable
class MultiTestcenterConnector
{
    /**
     * @var TestcenterConnector[]
     */
    private $connectors = [];

    /**
     * @param TestcenterInstanceConfig[]|null $instanceConfigs
     * @param TestcenterConnector[]|null $testcenterConnectors optional param to inject existing testcenter connectors
     */
    public function __construct(?array $instanceConfigs, array $testcenterConnectors = null)
    {
        $guzzleClient = new Client();
        if ($instanceConfigs !== null) {
            foreach ($instanceConfigs as $config) {
                $this->connectors[] = new TestcenterConnector(
                    $config->getUser(),
                    $config->getBaseUrl(),
                    null,
                    null,
                    null,
                    null,
                    $guzzleClient
                );
            }
        } elseif ($testcenterConnectors !== null) {
            $this->connectors = $testcenterConnectors;
        } else {
            throw new LogicException('Either $instanceConfigs or $testcenterConnectors must be set');
        }
    }

    /**
     * @throws TestcenterConnectorException
     */
    public function __call(string $method, array $args)
    {
        $results = [];
        $errors = [];

        if (preg_match('/.*Async$/', $method)) {
            $promises = [];

            foreach ($this->connectors as $connector) {
                if (method_exists($connector, $method)) {
                    $promises[$connector->getBaseUrl()] = call_user_func_array([$connector, $method], $args);
                } else {
                    throw new BadMethodCallException("Method $method does not exist in TestcenterConnector.");
                }
            }
            try {
                $results = Utils::unwrap($promises);
            } catch (Throwable $e) {
                throw new TestcenterConnectorException("Errors occurred: " . print_r($e, true));
            }
        } else {
            foreach ($this->connectors as $connector) {
                if (method_exists($connector, $method)) {
                    try {
                        $results[$connector->getBaseUrl()] = call_user_func_array([$connector, $method], $args);
                    } catch (Exception $e) {
                        $errors[$connector->getBaseUrl()] = "Error: " . $e->getMessage();
                    }
                } else {
                    throw new BadMethodCallException("Method $method does not exist in TestcenterConnector.");
                }
            }

            if (!empty($errors)) {
                throw new TestcenterConnectorException("Errors occurred: " . print_r($errors, true));
            }
        }

        return $results;
    }
}
