<?php

namespace Kompetenztestde\TestcenterConnector;

use Exception;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileGetResponse;
use GuzzleHttp\Client;
use GuzzleHttp\Promise\FulfilledPromise;
use GuzzleHttp\Promise\PromiseInterface;
use GuzzleHttp\Promise\Promise;
use GuzzleHttp\Promise\RejectedPromise;
use Kompetenztestde\TestcenterConnector\Api\Schemas\Result;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SessionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SysCheckReport;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SystemConfigResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UnitResponse;
use Kompetenztestde\TestcenterConnector\Models\TestGroup;
use Kompetenztestde\TestcenterConnector\Models\Workspace;
use LogicException;
use Throwable;
use ZipArchive;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileDeletionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileUploadResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\LogEntry;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserRole;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserWorkspace;
use Kompetenztestde\TestcenterConnector\Api\TestcenterApiClient;
use Kompetenztestde\TestcenterConnector\Exceptions\ApiException;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;
use Kompetenztestde\TestcenterConnector\Models\File;
use Kompetenztestde\TestcenterConnector\Models\User;
use Kompetenztestde\TestcenterConnector\Xml\XmlGenerator;

class TestcenterConnector
{
    /**
     * @var TestcenterApiClient
     */
    private $apiClient;

    /**
     * @var User
     */
    private $user;

    /**
     * @var string|null
     */
    private $authToken = null;

    /**
     * @var string|null
     */
    private $authenticatedUsername = null;

    /**
     * @var callable|null
     */
    private $authTokenChangedCallback;

    /*
     * @var int|null
     */
    private $userId = null;

    /**
     * Create a TestcenterConnector instance
     *
     * @param User $user user to authenticate with the testcenter instance
     * @param string|null $baseUrl base url of the testcenter instance (required if no api client is given)
     * @param TestcenterApiClient|null $apiClient optional parameter to inject api client instance
     * @param string|null $authToken preset a cached authentication token
     * @param string|null $authenticatedUsername preset a cached authenticated username
     */
    public function __construct(
        User $user,
        string $baseUrl = null,
        TestcenterApiClient $apiClient = null,
        string $authToken = null,
        string $authenticatedUsername = null,
        callable $authTokenChangedCallback = null,
        Client $guzzleClient = null
    ) {
        if ($apiClient === null && $baseUrl === null) {
            throw new LogicException('Either apiClient or baseUrl arguments are needed.');
        }

        if ($apiClient === null) {
            $this->apiClient = new TestcenterApiClient($baseUrl, $guzzleClient);
        } else {
            $this->apiClient = $apiClient;
        }

        $this->user = $user;
        $this->authToken = $authToken;
        $this->authenticatedUsername = $authenticatedUsername;
        $this->authTokenChangedCallback = $authTokenChangedCallback;
    }

    /**
     * Create a new group of testtakers and monitor logins for a booklet
     *
     * @param int|string $workspace ID or name of the workspace to create the group in
     * @param TestGroup $group group to create
     * @param string $bookletId ID of the booklet the group should be assigned to
     * @param XmlGenerator|null $xmlGenerator optional xml generator instance (for testing purposes)
     * @return FileUploadResponse file validation result of the testcenter
     * @throws TestcenterConnectorException
     */
    public function addTestGroup(
        $workspace,
        Models\TestGroup $group,
        string $bookletId,
        XmlGenerator $xmlGenerator = null
    ): FileUploadResponse {
        return $this->addTestGroupAsync($workspace, $group, $bookletId, $xmlGenerator)->wait();
    }

    /**
     * Asynchronously create a new group of testtakers and monitor logins for a booklet
     *
     * @param int|string $workspace ID or name of the workspace to create the group in
     * @param TestGroup $group group to create
     * @param string $bookletId ID of the booklet the group should be assigned to
     * @param XmlGenerator|null $xmlGenerator optional xml generator instance (for testing purposes)
     * @return PromiseInterface resolves to {@see FileUploadResponse} or rejects with
     *                          {@see TestcenterConnectorException}
     */
    public function addTestGroupAsync(
        $workspace,
        Models\TestGroup $group,
        string $bookletId,
        XmlGenerator $xmlGenerator = null
    ): PromiseInterface {
        $generatedTesttakers = null;

        if ($xmlGenerator === null) {
            $xmlGenerator = new XmlGenerator();
        }

        return $this->apiClient->getSystemConfigAsync()
            ->then(
                function (
                    SystemConfigResponse $config
                ) use (
                    $group,
                    $bookletId,
                    $xmlGenerator,
                    $workspace,
                    &$generatedTesttakers
                ) {
                    $version = $config->getVersion() ?? '15.1.4';

                    // generate testtakers xml
                    $xmlGroup = $this->getXmlTestGroup($group, $bookletId);
                    $generatedTesttakers = $xmlGenerator->generateTesttakersXml($xmlGroup, $version);

                    // upload testtakers xml
                    return $this->setAuthTokenAsync();
                }
            )
            ->then(function () use ($workspace, $group, &$generatedTesttakers) {
                if ($generatedTesttakers === null) {
                    throw new TestcenterConnectorException('Failed to generate testtakers XML');
                }
                $workspaceId = $this->getWorkspaceId($workspace);
                return $this->apiClient->uploadFileAsync(
                    $this->authToken,
                    $workspaceId,
                    'testtakers-' . $group->getId() . '.xml',
                    $generatedTesttakers
                );
            })
            ->then(function (FileUploadResponse $response) {
                return $response;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Create a new workspace
     *
     * @param string $name Name of the new workspace (must be unique)
     * @param bool $setRole set RW role for the current user for the new workspace (default: false)
     * @return int id of the created workspace
     * @throws TestcenterConnectorException
     */
    public function addWorkspace(string $name, bool $setRole = false): int
    {
        return $this->addWorkspaceAsync($name, $setRole)->wait();
    }

    /**
     * Asynchronously create a new workspace
     *
     * @param string $name Name of the new workspace (must be unique)
     * @param bool $setRole set RW role for the current user for the new workspace (default: false)
     * @return PromiseInterface resolves to workspace id (int) or rejects with {@see TestcenterConnectorException}
     */
    public function addWorkspaceAsync(string $name, bool $setRole = false): PromiseInterface
    {
        $createdWorkspaceId = null;
        return $this->setAuthTokenAsync()
            ->then(function () use ($name) {
                return $this->apiClient->addAWorkspaceAsync($this->authToken, $name);
            })
            ->then(function ($workspaceId) use ($name) {
                if ($workspaceId === null) {
                    return $this->getWorkspaceIdAsync($name);
                }
                return $workspaceId;
            })
            ->then(function ($workspaceId) use ($setRole, &$createdWorkspaceId) {
                $createdWorkspaceId = $workspaceId;
                if ($setRole) {
                    return $this->userId ?? $this->getUserIdAsync($this->user->getUsername());
                }
                return true;
            })
            ->then(function ($userId) use (&$createdWorkspaceId, $setRole) {
                if ($setRole) {
                    if ($createdWorkspaceId === null) {
                        throw new TestcenterConnectorException('WorkspaceId is null');
                    }
                    return $this->apiClient->changeUserRolesAsync(
                        $this->authToken,
                        $createdWorkspaceId,
                        [new UserRole($userId, UserRole::RW)]
                    );
                }
                return true;
            })
            ->then(function ($success) use (&$createdWorkspaceId) {
                if (!$success) {
                    throw new TestcenterConnectorException(
                        'Failed to set RW role for new workspace for current user.'
                    );
                }

                return $createdWorkspaceId;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Get a list of all workspaces
     *
     * @return Models\Workspace[]
     * @throws Exceptions\TestcenterConnectorException
     */
    public function getWorkspaces(): array
    {
        return $this->getWorkspacesAsync()->wait();
    }

    /**
     * Asynchronously get a list of all workspaces
     *
     * @return PromiseInterface resolves to array of {@see Workspace} on success or rejects with
     *                          {@see TestcenterConnectorException}
     */
    public function getWorkspacesAsync(): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () {
                return $this->apiClient->getAListOfWorkspacesAsync($this->authToken);
            })
            ->then(function ($workspaces) {
                return array_map(function ($workspace) {
                    return new Models\Workspace($workspace->getId(), $workspace->getName());
                }, $workspaces);
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Delete workspaces
     *
     * @param int|string[] $workspaces array of workspace ids or names to delete
     * @return bool
     * @throws TestcenterConnectorException
     */
    public function deleteWorkspace(array $workspaces): bool
    {
        return $this->deleteWorkspaceAsync($workspaces)->wait();
    }

    /**
     * Asynchronously delete workspaces
     *
     * @param int|string[] $workspaces array of workspace ids or names to delete
     * @return PromiseInterface resolves to true on success or rejects with {@see TestcenterConnectorException}
     */
    public function deleteWorkspaceAsync(array $workspaces): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () use ($workspaces) {
                return $this->getWorkspaceIdsAsync($workspaces);
            })
            ->then(function ($workspaceIds) {
                return $this->apiClient->deleteSomeWorkspacesAsync($this->authToken, $workspaceIds);
            })
            ->then(function ($success) {
                return $success;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Add a new test in the testcenter by uploading the required resources
     *
     * @param int|string $workspace ID or name of the workspace to create the test in
     * @param Models\TestInterface $test test definition
     * @return FileUploadResponse file validation result of the testcenter
     * @throws TestcenterConnectorException
     */
    public function addTest($workspace, Models\TestInterface $test): FileUploadResponse
    {
        $tmpFile = tmpfile();
        if ($tmpFile === false) {
            throw new TestcenterConnectorException('Add new Test: Failed to create temporary file');
        }
        $tmpLocation = stream_get_meta_data($tmpFile)['uri'];

        $zip = new ZipArchive();
        if ($zip->open($tmpLocation, ZipArchive::OVERWRITE) !== true) {
            fclose($tmpFile);
            throw new TestcenterConnectorException('Add new Test: Failed to create zip archive');
        }

        foreach ($test->getResources() as $filename => $content) {
            $zip->addFromString($filename, $content);
        }

        foreach ($test->getUnits() as $filename => $content) {
            $zip->addFromString($filename, $content);
        }

        $zip->addFromString($test->getBookletFilename(), $test->getBookletContent());

        if (!$zip->close()) {
            fclose($tmpFile);
            throw new TestcenterConnectorException('Add new Test: Failed to close zip archive');
        }

        $zipContent = file_get_contents($tmpLocation);
        if ($zipContent === false) {
            fclose($tmpFile);
            throw new TestcenterConnectorException('Add new Test: Failed to read temporary file');
        }

        fclose($tmpFile);

        try {
            $this->setAuthToken();
            $workspaceId = $this->getWorkspaceId($workspace);
            return $this->apiClient->uploadFile($this->authToken, $workspaceId, 'test.zip', $zipContent);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Get results of a single group
     *
     * @param int|string $workspace id or name of the workspace to get results from
     * @param string $groupId id of the test group to get results from
     * @return UnitResponse[]
     * @throws TestcenterConnectorException
     */
    public function getGroupResults($workspace, string $groupId): array
    {
        return $this->getGroupResultsAsync($workspace, $groupId)->wait();
    }

    /**
     * Asynchronously get results of a single group
     *
     * @param int|string $workspace id or name of the workspace to get results from
     * @param string $groupId id of the test group to get results from
     * @return PromiseInterface resolves to array of {@see UnitResponse} on success or rejects with
     *                          {@see TestcenterConnectorException}
     */
    public function getGroupResultsAsync($workspace, string $groupId): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () use ($workspace) {
                return $this->getWorkspaceIdAsync($workspace);
            })
            ->then(function ($workspaceId) use ($groupId) {
                return $this->apiClient->getReportOfItemResponsesAsync($this->authToken, $workspaceId, [$groupId]);
            })
            ->then(function ($responses) {
                return $responses;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    private function getXmlTestGroup(Models\TestGroup $group, string $bookletId): Xml\TestGroup
    {
        $xmlLogins = array_map(function ($login) use ($group) {
            $loginMode = ($login instanceof Models\Teacher)
                ? Xml\LoginMode::MONITOR_GROUP : Xml\LoginMode::RUN_HOT_RETURN;

            // Combine username and password if group password is set to allow login by link
            if ($group->getPassword()) {
                $username = $group->getPassword() . '-' . $login->getUsername();
                $password = null;
            } else {
                $username = $login->getUsername();
                $password = $login->getPassword();
            }

            return new Xml\TestLogin($username, $password, $loginMode);
        }, $group->getLogins());

        return new Xml\TestGroup(
            $group->getId(),
            $group->getLabel(),
            $xmlLogins,
            [$bookletId],
            $group->getValidFrom(),
            $group->getValidTo()
        );
    }

    /**
     * Deletes a list of given user-ids. ids wich did not exist get skipped
     *
     * @param int|string[] $users ids or usernames of users to delete
     * @return bool
     * @throws TestcenterConnectorException
     */
    public function deleteUsers(array $users): bool
    {
        try {
            $this->setAuthToken();
            $userIds = $this->getUserIds($users);
            return $this->apiClient->deleteSomeUsers($this->authToken, $userIds);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Adds an user
     *
     * @param string $username name of the new user
     * @param string $password password of the new user
     * @return bool
     * @throws TestcenterConnectorException
     */
    public function addUser(string $username, string $password): bool
    {
        try {
            $this->setAuthToken();
            return $this->apiClient->addAUser($this->authToken, $username, $password);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Create a new user with super admin role
     *
     * @param string $username name of the new user
     * @param string $password password of the new user
     * @return int id of created user, false if unsuccessful
     * @throws TestcenterConnectorException
     */
    public function addSuperAdmin(string $username, string $password): int
    {
        try {
            $this->setAuthToken();
            $success = $this->apiClient->addAUser($this->authToken, $username, $password);

            if (!$success) {
                throw new TestcenterConnectorException("Create user '$username' failed");
            }

            $users = $this->apiClient->getAListOfUsers($this->authToken);
            $users = array_filter(
                $users,
                function (\Kompetenztestde\TestcenterConnector\Api\Schemas\User $user) use ($username) {
                    return $user->getName() === $username;
                }
            );

            if (count($users) !== 1) {
                throw new TestcenterConnectorException("Create user '$username' failed");
            }

            $users = array_values($users);
            $user = $users[0];

            // set super admin role
            $success = $this->apiClient->changeSuperAdminStatus(
                $this->authToken,
                $user->getId(),
                true,
                $this->user->getPassword()
            );

            if (!$success) {
                throw new TestcenterConnectorException('Setting super admin role failed.');
            }

            // set RW roles for all existing workspaces
            $workspaces = $this->apiClient->getAListOfWorkspaces($this->authToken);
            foreach ($workspaces as $workspace) {
                $success = $this->apiClient->changeUserRoles($this->authToken, $workspace->getId(), [
                    new UserRole($user->getId(), UserRole::RW)
                ]);

                if (!$success) {
                    throw new TestcenterConnectorException("Failed to set RW role for workspace {$workspace->getId()}");
                }
            }

            return $user->getId();
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Returns info about all registered users in a workspace.
     *
     * @return User[]
     * @throws TestcenterConnectorException
     */
    public function getUsers(): array
    {
        try {
            $this->setAuthToken();
            return $this->apiClient->getAListOfUsers($this->authToken);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Renames a workspace with given id
     *
     * @param int|string $workspace id or name of workspace to rename
     * @param string $name new name
     * @return bool
     * @throws TestcenterConnectorException
     */
    public function renameWorkspace($workspace, string $name): bool
    {
        try {
            $this->setAuthToken();
            $workspaceId = $this->getWorkspaceId($workspace);
            return $this->apiClient->renameAWorkspace($this->authToken, $workspaceId, $name);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Deletes all results and monitor data of a group of groups
     *
     * @param int|string $workspace id or name of the workspace to delete data from
     * @param string[] $groups group ids to delete data from
     * @return bool
     * @throws TestcenterConnectorException
     */
    public function deleteData($workspace, array $groups): bool
    {
        try {
            $this->setAuthToken();
            $workspaceId = $this->getWorkspaceId($workspace);
            return $this->apiClient->deleteData($this->authToken, $workspaceId, $groups);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Returns a System Check report in JSON or CSV format based on aggregated report data
     *
     * @param int|string $workspace id or name of the workspace
     * @param string[] $dataIds
     * @return SysCheckReport[]
     * @throws TestcenterConnectorException
     */
    public function getSystemCheckReport($workspace, array $dataIds): array
    {
        try {
            $this->setAuthToken();
            $workspaceId = $this->getWorkspaceId($workspace);
            $reports = $this->apiClient->getReportOfSystemChecks($this->authToken, $workspaceId, $dataIds);
            return $reports;
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Retrieves a list of unit- and booklet results for a given workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @return Result[]
     * @throws TestcenterConnectorException
     */
    public function getResults($workspace): array
    {
        return $this->getResultsAsync($workspace)->wait();
    }

    /**
     * Asynchronously retrieves a list of unit- and booklet results for a given workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @return PromiseInterface resolves to array of {@see Result} or rejects with {@see TestcenterConnectorException}
     */
    public function getResultsAsync($workspace): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () use ($workspace) {
                return $this->getWorkspaceIdAsync($workspace);
            })
            ->then(function (int $workspaceId) {
                return $this->apiClient->getResultsAsync($this->authToken, $workspaceId);
            })
            ->then(function ($results) {
                return $results;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Get an array of workspaces a user has access to
     *
     * @param int|string $user id or name of the user
     * @return UserWorkspace[]
     * @throws TestcenterConnectorException
     */
    public function getWorkspacesByUser($user): array
    {
        try {
            $this->setAuthToken();
            $userId = $this->getUserId($user);
            return $this->apiClient->getAListOfUserWorkspaces($this->authToken, $userId);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Set or change the roles of one or more users for a workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @param UserRole[] $userRoles array of roles to set/change
     * @return bool true if successfully changed roles
     * @throws TestcenterConnectorException
     */
    public function changeUserRoles($workspace, array $userRoles): bool
    {
        try {
            $this->setAuthToken();
            $workspaceId = $this->getWorkspaceId($workspace);
            return $this->apiClient->changeUserRoles($this->authToken, $workspaceId, $userRoles);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Get logs of a test group in a workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @param string[] $groupIds ids of the test groups to get logs from
     * @return LogEntry[]
     * @throws TestcenterConnectorException
     */
    public function getLogs($workspace, array $groupIds): array
    {
        return $this->getLogsAsync($workspace, $groupIds)->wait();
    }

    /**
     * Asynchronously get logs of a test group in a workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @param string[] $groupIds ids of the test groups to get logs from
     * @return PromiseInterface resolves to array of {@see LogEntry} or rejects with {@see TestcenterConnectorException}
     */
    public function getLogsAsync($workspace, array $groupIds): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () use ($workspace) {
                return $this->getWorkspaceIdAsync($workspace);
            })
            ->then(function (int $workspaceId) use ($groupIds) {
                return $this->apiClient->getReportOfLogsAsync($this->authToken, $workspaceId, $groupIds);
            })
            ->then(function ($logs) {
                return $logs;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Get information about files in a workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @return FileGetResponse
     * @throws TestcenterConnectorException
     */
    public function getWorkspaceFiles($workspace): FileGetResponse
    {
        return $this->getWorkspaceFilesAsync($workspace)->wait();
    }

    /**
     * Asynchronously get information about files in a workspace
     *
     * @param int|string $workspace id or name of the workspace
     * @return PromiseInterface resolves to {@see FileGetResponse} or rejects with {@see TestcenterConnectorException}
     */
    public function getWorkspaceFilesAsync($workspace): PromiseInterface
    {
        return $this->setAuthTokenAsync()
            ->then(function () use ($workspace) {
                return $this->getWorkspaceIdAsync($workspace);
            })
            ->then(function (int $workspaceId) {
                return $this->apiClient->getFilesOfWorkspaceAsync($this->authToken, $workspaceId);
            })
            ->then(function ($workspaceFiles) {
                return $workspaceFiles;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Delete one or more files in a workspace
     *
     * @param int|string $workspace id or name of the workspace to delete files from
     * @param File[] $files list of files to delete
     * @return FileDeletionResponse
     * @throws TestcenterConnectorException
     */
    public function deleteFiles($workspace, array $files): FileDeletionResponse
    {
        try {
            $this->setAuthToken();
            $filenames = array_map(function (File $file) {
                return $file->getType() . '/' . $file->getFilename();
            }, $files);
            $workspaceId = $this->getWorkspaceId($workspace);
            return $this->apiClient->deleteFiles($this->authToken, $workspaceId, $filenames);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Change the password of an existing user (if user does not exist, request gets ignored)
     *
     * @param int|string $user id or name of the user
     * @param string $newPassword
     * @return bool true if successfully changed
     * @throws TestcenterConnectorException
     */
    public function changeUserPassword($user, string $newPassword): bool
    {
        try {
            $this->setAuthToken();
            $userId = $this->getUserId($user);
            return $this->apiClient->changeUserPassword($this->authToken, $userId, $newPassword);
        } catch (ApiException $e) {
            throw new TestcenterConnectorException('API Exception: ' . $e);
        }
    }

    /**
     * Get the testcenter api base url of the connector instance
     *
     * @return string
     */
    public function getBaseUrl(): string
    {
        return $this->apiClient->getBaseUrl();
    }

    /**
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    private function setAuthToken(): void
    {
        $this->setAuthTokenAsync()->wait();
    }

    private function setAuthTokenAsync(): PromiseInterface
    {
        return $this->hasRunningSessionAsync()
            ->then(function (bool $result) {
                if (!$result) {
                    switch ($this->user->getRole()) {
                        case User::ADMIN_ROLE:
                            return $this->apiClient->startAdminSessionAsync(
                                $this->user->getUsername(),
                                $this->user->getPassword()
                            );
                        case User::LOGIN_ROLE:
                            throw new Exception("Not implemented");
                        default:
                            throw new TestcenterConnectorException(
                                "{$this->user->getRole()} is not an allowed user role."
                            );
                    }
                }
                return null;
            })
            ->then(function ($sessionResponse) {
                if (!$sessionResponse) {
                    return;
                }

                $this->authToken = $sessionResponse->getToken();
                $this->authenticatedUsername = $sessionResponse->getDisplayName();
                $this->userId = $sessionResponse->getId();

                if ($this->authTokenChangedCallback !== null) {
                    call_user_func($this->authTokenChangedCallback, $this->authToken);
                }
            });
    }

    private function hasRunningSessionAsync(): PromiseInterface
    {
        if ($this->authToken === null || $this->user->getUsername() !== $this->authenticatedUsername) {
            return new FulfilledPromise(false);
        }

        return $this->apiClient->getASessionAsync($this->authToken)
            ->then(
                function (SessionResponse $sessionResponse) {
                    return $sessionResponse->getToken() === $this->authToken;
                },
                function (ApiException $e) {
                    return false;
                }
            );
    }

    /**
     * Get workspace id by name or id
     *
     * @param int|string $workspace id or name of a workspace
     * @return int
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    private function getWorkspaceId($workspace): int
    {
        return $this->getWorkspaceIdAsync($workspace)->wait();
    }

    private function getWorkspaceIdAsync($workspace): PromiseInterface
    {
        if (is_string($workspace)) {
            return $this->apiClient->getAListOfWorkspacesAsync($this->authToken)
                ->then(function ($workspaces) use ($workspace) {
                    foreach ($workspaces as $currentWorkspace) {
                        if ($currentWorkspace->getName() === $workspace) {
                            return $currentWorkspace->getId();
                        }
                    }

                    throw new TestcenterConnectorException("Workspace with name '$workspace' does not exist.");
                });
        } elseif (is_int($workspace)) {
            return new FulfilledPromise($workspace);
        } else {
            return new Promise(function () {
                throw new LogicException('workspace must be either int or string');
            });
        }
    }

    /**
     * Get ids for an array of workspaces
     *
     * @param int|string[] $workspaces ids or names of workspaces
     * @return PromiseInterface resolves to array of workspace ids (int) or rejects with
     *                          {@see TestcenterConnectorException} or {@see LogicException}
     */
    private function getWorkspaceIdsAsync($workspaces): PromiseInterface
    {
        if (!$this->containsWorkspaceName($workspaces)) {
            return new FulfilledPromise($workspaces);
        }

        return $this->apiClient->getAListOfWorkspacesAsync($this->authToken)
            ->then(function ($instanceWorkspaces) use ($workspaces) {
                $workspaceIds = [];

                foreach ($workspaces as $workspace) {
                    if (is_string($workspace)) {
                        $workspaceId = null;
                        foreach ($instanceWorkspaces as $instanceWorkspace) {
                            if ($instanceWorkspace->getName() === $workspace) {
                                $workspaceId = $instanceWorkspace->getId();
                                break;
                            }
                        }

                        if ($workspaceId !== null) {
                            $workspaceIds[] = $workspaceId;
                        } else {
                            throw new TestcenterConnectorException("Workspace with name '$workspace' does not exist.");
                        }
                    } elseif (is_int($workspace)) {
                        $workspaceIds[] = $workspace;
                    } else {
                        throw new LogicException('workspace must be either int or string');
                    }
                }

                return $workspaceIds;
            }, function ($e) {
                $this->handleAsyncError($e);
            });
    }

    /**
     * Check if an array of workspaces contains a workspace by name
     *
     * @param int|string[] $workspaces array of workspace ids or names
     * @return bool true if at least one instance is not an integer, false if all are integers
     */
    private function containsWorkspaceName($workspaces): bool
    {
        foreach ($workspaces as $workspace) {
            if (!is_int($workspace)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Get a users id
     *
     * @param int|string $user id or name of a user
     * @return int
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    private function getUserId($user): int
    {
        return $this->getUserIdAsync($user)->wait();
    }

    /**
     * Asynchronously get a users id
     *
     * @param int|string $user id or name of a user
     * @return PromiseInterface resolves to user id (int) or rejects with {@see TestcenterConnectorException} or
     *                          {@see LogicException}
     */
    private function getUserIdAsync($user): PromiseInterface
    {
        if (is_string($user)) {
            return $this->apiClient->getAListOfUsersAsync($this->authToken)
                ->then(function ($users) use ($user) {
                    foreach ($users as $currentUser) {
                        if ($currentUser->getName() === $user) {
                            return $currentUser->getId();
                        }
                    }
                    return null;
                })
                ->then(function ($userId) use ($user) {
                    if ($userId === null) {
                        throw new TestcenterConnectorException("User with name '$user' does not exist.");
                    }
                    return $userId;
                });
        } elseif (is_int($user)) {
            return new FulfilledPromise($user);
        } else {
            return new RejectedPromise(new LogicException('user must be either int or string'));
        }
    }

    /**
     * Get ids for an array of users
     *
     * @param int|string[] $users array of ids or names of users
     * @return int[]
     * @throws ApiException
     * @throws TestcenterConnectorException
     */
    private function getUserIds($users): array
    {
        $instanceUsers = null;
        $userIds = [];

        foreach ($users as $user) {
            if (is_string($user)) {
                if ($instanceUsers === null) {
                    $instanceUsers = $this->apiClient->getAListOfUsers($this->authToken);
                }

                $userId = null;
                foreach ($instanceUsers as $currentUser) {
                    if ($currentUser->getName() === $user) {
                        $userId = $currentUser->getId();
                        break;
                    }
                }

                if ($userId !== null) {
                    $userIds[] = $userId;
                } else {
                    throw new TestcenterConnectorException("User with name '$user' does not exist.");
                }
            } elseif (is_int($user)) {
                $userIds[] = $user;
            } else {
                throw new LogicException('user must be either int or string');
            }
        }

        return $userIds;
    }

    /**
     * Handle an error received in a promise's then block
     *
     * @throws Throwable
     * @throws TestcenterConnectorException
     */
    private function handleAsyncError($error)
    {
        if ($error instanceof ApiException) {
            throw new TestcenterConnectorException('API Exception: ' . $error);
        } elseif ($error instanceof Throwable) {
            throw $error;
        } else {
            throw new Exception($error);
        }
    }
}
