<?php

namespace Kompetenztestde\TestcenterConnector\Api;

use DateTime;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Handler\MockHandler;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Middleware;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Psr7\Response;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserRole;
use Kompetenztestde\TestcenterConnector\Exceptions\ApiException;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;
use PHPUnit\Framework\TestCase;

class TestcenterApiClientTest extends TestCase
{
    /**
     * @var Client
     */
    private $httpClient;

    /**
     * @var MockHandler
     */
    private $mock;

    /**
     * @var TestcenterApiClient
     */
    private $apiClient;

    /**
     * @var array
     */
    private $container;

    protected function setUp(): void
    {
        $this->mock = new MockHandler([]);
        $this->container = [];
        $history = Middleware::history($this->container);
        $handlerStack = HandlerStack::create($this->mock);
        $handlerStack->push($history);
        $this->httpClient = new Client(['handler' => $handlerStack]);

        $this->apiClient = new TestcenterApiClient('http://localhost/api', $this->httpClient);
    }

    public function testStartAdminSession(): void
    {
        $responseBody = '{
            "token": "user000000000.0000000000",
            "displayName": "super",
            "access": {
              "workspaceAdmin": [
                1
              ],
              "superAdmin": []
            }
        }';
        $this->mock->append(new Response(200, [], $responseBody));

        $response = $this->apiClient->startAdminSession('super', 'user123');
        $this->assertEquals('user000000000.0000000000', $response->getToken());
    }

    public function testStartAdminSessionClientError(): void
    {
        $this->mock->append(new Response(400, [], ''));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(400);
        $this->apiClient->startAdminSession('super', 'user123');
    }

    public function testStartAdminSessionNoAdminPriv(): void
    {
        $this->mock->append(new Response(204, [], ''));
        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('User exists but has no admin privileges.');
        $this->apiClient->startAdminSession('super', 'user123');
    }

    public function testStartAdminSessionServerError(): void
    {
        $this->mock->append(new Response(500, [], ''));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->startAdminSession('super', 'user123');
    }

    public function testStartAdminSessionInvalidResponse(): void
    {
        $this->mock->append(new Response(200, [], '{}'));
        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "token" is missing in response');
        $this->apiClient->startAdminSession('super', 'user123');
    }

    public function testStartAdminSessionConnectionError(): void
    {
        $this->mock->append(new ConnectException('Failed to connect to server', new Request('POST', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessageMatches('/Failed to connect to testcenter server \(Reason: .*\)/');
        $this->apiClient->startAdminSession('super', 'user123');
    }

    public function testUploadFile(): void
    {
        $responseBody = '{
            "SAMPLE_UNIT.XML": {
                "warning": [
                    "File of name `SAMPLE_UNIT.XML` did already exist and was overwritten."
                ]
            }
        }';
        $this->mock->append(new Response(201, [], $responseBody));

        $response = $this->apiClient->uploadFile('test-token', 1, 'SAMPLE_UNIT.XML', 'test');

        $this->assertEquals(1, count($response->getFileValidationResults()));
        $this->assertTrue($response->hasWarnings());
        $this->assertFalse($response->hasErrors());

        $result = $response->getFileValidationResults()[0];
        $this->assertEquals('SAMPLE_UNIT.XML', $result->getFilename());
        $this->assertEquals(1, count($result->getWarnings()));
        $this->assertEquals(0, count($result->getErrors()));

        $warning = $result->getWarnings()[0];
        $this->assertEquals(
            "File of name `SAMPLE_UNIT.XML` did already exist and was overwritten.",
            $warning
        );
    }

    public function testUploadFileError(): void
    {
        $responseBody = '{
            "SAMPLE_UNIT.XML": {
              "error": [
                "Invalid root-tag: `Invalid`"
              ]
            }
        }';
        $this->mock->append(new Response(207, [], $responseBody));

        $response = $this->apiClient->uploadFile('test-token', 1, 'SAMPLE_UNIT.XML', 'test');
        $this->assertTrue($response->hasErrors());
    }

    /**
     * @dataProvider uploadFileClientErrorProvider
     */
    public function testUploadFileClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode, [], ''));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->uploadFile('test-token', 1, 'SAMPLE_UNIT.XML', 'test');
    }

    public function testUploadFileServerError(): void
    {
        $this->mock->append(new Response(500, [], ''));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->uploadFile('test-token', 1, 'SAMPLE_UNIT.XML', 'test');
    }

    public function testUploadFileConnectionError(): void
    {
        $this->mock->append(new ConnectException('Failed to connect to server', new Request('POST', '/test')));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessageMatches('/Failed to connect to testcenter server \(Reason: .*\)/');
        $this->apiClient->uploadFile('test-token', 1, 'SAMPLE_UNIT.XML', 'test');
    }

    public static function uploadFileClientErrorProvider(): array
    {
        return [
            [400],
            [401],
            [403],
            [404],
            [409],
            [410],
            [413]
        ];
    }

    public function testDeleteFiles(): void
    {
        $body = file_get_contents('testAssets/responses/delete_files.json');
        $this->mock->append(new Response(207, [], $body));

        $response = $this->apiClient->deleteFiles('test-token', 1, [
            'SysCheck/SAMPLE_SYSCHECK.XML',
            'Unit/SAMPLE_UNIT.XML',
            'some rubbish',
            'a/b',
            '../../README.md'
        ]);

        $this->assertCount(1, $response->getDeleted());
        $this->assertContains('SysCheck/SAMPLE_SYSCHECK.XML', $response->getDeleted());
        $this->assertCount(2, $response->getDidNotExist());
        $this->assertContains('some rubbish', $response->getDidNotExist());
        $this->assertContains('a/b', $response->getDidNotExist());
        $this->assertCount(1, $response->getNotAllowed());
        $this->assertContains('../../README.md', $response->getNotAllowed());
        $this->assertCount(1, $response->getWasUsed());
        $this->assertContains('Unit/SAMPLE_UNIT.XML', $response->getWasUsed());
    }

    public function testAddAWorkspace(): void
    {
        $this->mock->append(new Response(201, [], '2'));
        $workspaceId = $this->apiClient->addAWorkspace('test-token', 'test-workspace');
        $this->assertEquals(1, count($this->container));

        $transaction = $this->container[0];

        $this->assertEquals('PUT', $transaction['request']->getMethod());
        $this->assertEquals('/api/workspace', $transaction['request']->getUri()->getPath());

        $requestBody = $transaction['request']->getBody()->getContents();
        $requestData = json_decode($requestBody);
        $this->assertEquals('test-workspace', $requestData->name);

        $this->assertEquals(2, $workspaceId);
    }

    public function testAddAWorkspaceLegacy(): void
    {
        $this->mock->append(new Response(201, [], ''));
        $response = $this->apiClient->addAWorkspace('test-token', 'test-workspace');
        $this->assertEquals(1, count($this->container));

        $transaction = $this->container[0];

        $this->assertEquals('PUT', $transaction['request']->getMethod());
        $this->assertEquals('/api/workspace', $transaction['request']->getUri()->getPath());

        $requestBody = $transaction['request']->getBody()->getContents();
        $requestData = json_decode($requestBody);
        $this->assertEquals('test-workspace', $requestData->name);

        $this->assertNull($response);
    }

    /**
     * @dataProvider addAWorkspaceClientErrorProvider
     */
    public function testAddAWorkspaceClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->addAWorkspace('test-token', 'test-workspace');
    }

    public function testAddAWorkspaceServerError(): void
    {
        $this->mock->append(new Response(500));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->addAWorkspace('test-token', 'test-workspace');
    }

    public function testAddAWorkspaceConnectionError(): void
    {
        $this->mock->append(new ConnectException('Connection error', new Request('PUT', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessageMatches('/Failed to connect to testcenter server \(Reason: .*\)/');
        $this->apiClient->addAWorkspace('test-token', 'test-workspace');
    }

    public static function addAWorkspaceClientErrorProvider(): array
    {
        return [
            [401],
            [410]
        ];
    }

    public function testGetAListOfWorkspaces(): void
    {
        $responseBody = '[{
            "id": 1,
            "name": "example_workspace_1"
        }, {
            "id": 2,
            "name": "example_workspace_2"
        }]';
        $this->mock->append(new Response(200, [], $responseBody));

        $workspaces = $this->apiClient->getAListOfWorkspaces('test-token');
        $this->assertEquals(2, count($workspaces));

        $this->assertEquals(1, $workspaces[0]->getId());
        $this->assertEquals('example_workspace_1', $workspaces[0]->getName());

        $this->assertEquals(2, $workspaces[1]->getId());
        $this->assertEquals('example_workspace_2', $workspaces[1]->getName());
    }

    /**
     * @dataProvider getAListOfWorkspacesClientErrorProvider
     */
    public function testGetAListOfWorkspacesClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode, [], ''));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesServerError(): void
    {
        $this->mock->append(new Response(500, [], ''));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesConnectionError(): void
    {
        $this->mock->append(new ConnectException('Connection Failed', new Request('GET', '/test')));
        $this->expectExceptionMessageMatches('/Failed to connect to testcenter server \(Reason: .*\)/');
        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionCode(0);
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesNoId(): void
    {
        $responseBody = '[{
            "id": 1,
            "name": "example_workspace_1"
        }, {
            "name": "example_workspace_2"
        }]';
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "id" is missing in response');
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesInvalidId(): void
    {
        $responseBody = '[{
            "id": 1,
            "name": "example_workspace_1"
        }, {
            "id": "Test",
            "name": "example_workspace_2"
        }]';
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "id" expected to be of type integer. Actual: string');
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesNoName(): void
    {
        $responseBody = '[{
            "id": 1,
            "name": "example_workspace_1"
        }, {
            "id": 2
        }]';
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "name" is missing in response');
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public function testGetAListOfWorkspacesInvalidName(): void
    {
        $responseBody = '[{
            "id": 1,
            "name": "example_workspace_1"
        }, {
            "id": 2,
            "name": 2
        }]';
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "name" expected to be of type string. Actual: integer');
        $this->apiClient->getAListOfWorkspaces('test-token');
    }

    public static function getAListOfWorkspacesClientErrorProvider(): array
    {
        return [
            [401],
            [403],
            [410],
        ];
    }

    public function testGetAListOfUserWorkspaces(): void
    {
        $this->mock->append(new Response(200, [], '[
            {
                "id": 1,
                "name": "example_workspace",
                "selected": false,
                "role": ""
            }
        ]'));

        $workspaces = $this->apiClient->getAListOfUserWorkspaces('test-token', 1);

        $this->assertCount(1, $workspaces);
        $this->assertEquals(1, $workspaces[0]->getId());
        $this->assertEquals('example_workspace', $workspaces[0]->getName());
        $this->assertFalse($workspaces[0]->isSelected());
    }

    /**
     * @dataProvider getAListOfUserWorkspacesErrorProvider
     */
    public function testGetAListOfUserWorkspacesError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->getAListOfUserWorkspaces('test-token', 1);
    }

    public function testGetAListOfUserWorkspacesConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('GET', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->getAListOfUserWorkspaces('test-token', 1);
    }

    public static function getAListOfUserWorkspacesErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }

    public function testChangeUserRoles(): void
    {
        $this->mock->append(new Response(200));
        $roles = [
            new UserRole(1, UserRole::RW),
            new UserRole(2, UserRole::RO)
        ];

        $this->apiClient->changeUserRoles('test-token', 1, $roles);

        $this->assertEquals(1, count($this->container));

        $request = $this->container[0]['request'];
        $this->assertEquals('PATCH', $request->getMethod());
        $this->assertEquals('/api/workspace/1/users', $request->getUri()->getPath());

        $body = $request->getBody()->getContents();
        $data = json_decode($body);
        $this->assertTrue(isset($data->u));
        $this->assertEquals(2, count($data->u));
        $this->assertEquals(1, $data->u[0]->id);
        $this->assertEquals('RW', $data->u[0]->role);
        $this->assertEquals(2, $data->u[1]->id);
        $this->assertEquals('RO', $data->u[1]->role);
    }

    /**
     * @dataProvider getAListOfWorkspacesClientErrorProvider
     */
    public function testChangeUserRolesClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->changeUserRoles('test-token', 1, [
            new UserRole(1, UserRole::RW)
        ]);
    }

    public function testChangeUserRolesServerError(): void
    {
        $this->mock->append(new Response(500));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->changeUserRoles('test-token', 1, [
            new UserRole(1, UserRole::RW)
        ]);
    }

    public function testChangeUserRolesConnectionError(): void
    {
        $this->mock->append(new ConnectException('Connection error', new Request('PATCH', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->apiClient->changeUserRoles('test-token', 1, [
            new UserRole(1, UserRole::RW)
        ]);
    }

    public static function changeUserRolesClientErrorProvider(): array
    {
        return [
            [401],
            [403],
            [410]
        ];
    }

    public function testGetReportOfItemResponses(): void
    {
        $responseBody = '[
            {
              "groupname": "sample_group",
              "loginname": "test",
              "code": "xxx",
              "bookletname": "BOOKLET.SAMPLE-1",
              "unitname": "UNIT.SAMPLE",
              "laststate": "{\"PLAYER\":\"RUNNING\",\"CURRENT_PAGE_NR\":\"21\",\"CURRENT_PAGE_ID\":\"2\",\"PAGE_COUNT\":\"4\",\"PRESENTATION_PROGRESS\":\"some\",\"RESPONSE_PROGRESS\":\"some\"}",
              "responses": [
                {
                  "id": "answers",
                  "content": "{\"required-text-field\":\"test\",\"readonly-text-field\":\"read-only\",\"text-field\":\"test123\",\"number-field\":\"\",\"date-field\":\"\",\"email-field\":\"\",\"range-field\":\"6\",\"select-box\":\"a\",\"list-field\":\"\",\"text-area\":\"Type something...\",\"\":[\"a\",\"b\",\"c\",\"d\"],\"editable-named-required\":\"\"}",
                  "ts": 1713855978063,
                  "responseType": "iqb-simple-player@2.0.0"
                },
                {
                  "id": "audio",
                  "content": "{\"soundOn\":false,\"position\":0,\"length\":20}",
                  "ts": 1713855970866,
                  "responseType": "iqb-simple-player@2.0.0"
                }
              ]
            },
            {
              "groupname": "sample_group",
              "loginname": "test",
              "code": "xxx",
              "bookletname": "BOOKLET.SAMPLE-1",
              "unitname": "UNIT.SAMPLE-2",
              "laststate": "{\"PLAYER\":\"LOADING\"}",
              "responses": []
            },
            {
              "groupname": "sample_group",
              "loginname": "test",
              "code": "xxx",
              "bookletname": "BOOKLET.SAMPLE-1",
              "unitname": "an_alias",
              "laststate": "{\"PLAYER\":\"RUNNING\",\"CURRENT_PAGE_NR\":\"11\",\"CURRENT_PAGE_ID\":\"1\",\"PAGE_COUNT\":\"4\",\"PRESENTATION_PROGRESS\":\"some\",\"RESPONSE_PROGRESS\":\"some\"}",
              "responses": [
                {
                  "id": "answers",
                  "content": "{\"required-text-field\":\"\",\"readonly-text-field\":\"read-only\",\"text-field\":\"\",\"number-field\":\"\",\"date-field\":\"\",\"email-field\":\"\",\"range-field\":\"6\",\"select-box\":\"a\",\"list-field\":\"\",\"text-area\":\"Type something...\",\"\":[\"a\",\"b\",\"c\",\"d\"],\"editable-named-required\":\"\"}",
                  "ts": 1713855985204,
                  "responseType": "iqb-simple-player@2.0.0"
                },
                {
                  "id": "audio",
                  "content": "{\"soundOn\":false,\"position\":0,\"length\":20}",
                  "ts": 1713855985204,
                  "responseType": "iqb-simple-player@2.0.0"
                }
              ]
            }
        ]';
        $this->mock->append(new Response(200, [], $responseBody));
        $responses = $this->apiClient->getReportOfItemResponses('test-token', 1, ['sample_group']);
        $this->assertEquals(3, count($responses));

        foreach ($responses as $response) {
            $this->assertEquals('sample_group', $response->getGroupName());
            $this->assertEquals('test', $response->getLoginName());
            $this->assertEquals('xxx', $response->getCode());
            $this->assertEquals('BOOKLET.SAMPLE-1', $response->getBookletName());
        }

        $response1 = $responses[0];
        $this->assertEquals(2, count($response1->getResponses()));
        $this->assertEquals('UNIT.SAMPLE', $response1->getUnitName());
        $this->assertEquals('answers', $response1->getResponses()[0]->getId());
        $this->assertEquals(1713855978063, $response1->getResponses()[0]->getTs());
        $this->assertEquals('iqb-simple-player@2.0.0', $response1->getResponses()[0]->getResponseType());
    }

    public function testGetReportOfLogs(): void
    {
        $responseBody = '[
            {
                "groupname": "sample_group",
                "loginname": "test",
                "code": "xxx",
                "bookletname": "BOOKLET.SAMPLE-1",
                "unitname": "UNIT.SAMPLE",
                "timestamp": 1713855968923,
                "logentry": "PLAYER = LOADING"
            },
            {
                "groupname": "sample_group",
                "loginname": "test",
                "code": "xxx",
                "bookletname": "BOOKLET.SAMPLE-1",
                "unitname": "UNIT.SAMPLE",
                "timestamp": 1713855968968,
                "logentry": "PLAYER = RUNNING"
            }
        ]';

        $this->mock->append(new Response(200, [], $responseBody));
        $logEntries = $this->apiClient->getReportOfLogs('test-token', 1, ['sample_group']);

        $this->assertEquals(2, count($logEntries));

        foreach ($logEntries as $logEntry) {
            $this->assertEquals('sample_group', $logEntry->getGroupName());
            $this->assertEquals('test', $logEntry->getLoginName());
            $this->assertEquals('xxx', $logEntry->getCode());
            $this->assertEquals('BOOKLET.SAMPLE-1', $logEntry->getBookletName());
            $this->assertEquals('UNIT.SAMPLE', $logEntry->getUnitName());
        }

        $this->assertEquals(1713855968923, $logEntries[0]->getTimestamp());
        $this->assertEquals('PLAYER = LOADING', $logEntries[0]->getLogEntry());

        $this->assertEquals(1713855968968, $logEntries[1]->getTimestamp());
        $this->assertEquals('PLAYER = RUNNING', $logEntries[1]->getLogEntry());
    }

    public function testGetASession(): void
    {
        $response = '{
            "token": "static:person:sample_group_test_xxx",
            "displayName": "sample_group/test/xxx",
            "claims": {
                "test": [
                    {
                        "label": "Reduced Booklet",
                        "id": "BOOKLET.SAMPLE-2",
                        "type": "test",
                        "flags": {
                            "locked": true,
                            "running": true
                        }
                    }
                ]
            },
            "access": {
                "test": [
                    "BOOKLET.SAMPLE-1"
                ]
            },
            "customTexts": {},
            "flags": []
        }';

        $this->mock->append(new Response(200, [], $response));

        $response = $this->apiClient->getASession('test-token');
        $this->assertEquals('static:person:sample_group_test_xxx', $response->getToken());
        $this->assertEquals('sample_group/test/xxx', $response->getDisplayName());
    }

    public function testGetASessionNoToken(): void
    {
        $response = '{
            "displayName": "sample_group/test/xxx",
            "claims": {
                "test": [
                    {
                        "label": "Reduced Booklet",
                        "id": "BOOKLET.SAMPLE-2",
                        "type": "test",
                        "flags": {
                            "locked": true,
                            "running": true
                        }
                    }
                ]
            },
            "access": {
                "test": [
                    "BOOKLET.SAMPLE-1"
                ]
            },
            "customTexts": {},
            "flags": []
        }';

        $this->mock->append(new Response(200, [], $response));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "token" is missing in response');
        $this->apiClient->getASession('test-token');
    }

    public function testGetASessionNoDisplayName(): void
    {
        $response = '{
            "token": "static:person:sample_group_test_xxx",
            "claims": {
                "test": [
                    {
                        "label": "Reduced Booklet",
                        "id": "BOOKLET.SAMPLE-2",
                        "type": "test",
                        "flags": {
                            "locked": true,
                            "running": true
                        }
                    }
                ]
            },
            "access": {
                "test": [
                    "BOOKLET.SAMPLE-1"
                ]
            },
            "customTexts": {},
            "flags": []
        }';

        $this->mock->append(new Response(200, [], $response));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "displayName" is missing in response');
        $this->apiClient->getASession('test-token');
    }

    /**
     * @dataProvider getASessionClientErrorProvider
     */
    public function testGetASessionClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode, [], ''));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->getASession('test-token');
    }

    public function testGetASessionServerError(): void
    {
        $this->mock->append(new Response(500, [], ''));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);
        $this->apiClient->getASession('test-token');
    }

    public function testGetASessionConnectError(): void
    {
        $this->mock->append(new ConnectException('Connection failed', new Request('GET', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->apiClient->getASession('test-token');
    }

    public static function getASessionClientErrorProvider(): array
    {
        return [
            [401],
            [410]
        ];
    }

    public function testAddAUser(): void
    {
        $this->mock->append(new Response(201));

        $this->apiClient->addAUser('test-token', 'user1', 'user123');

        $this->assertCount(1, $this->container);

        $transaction = $this->container[0];

        $this->assertEquals('PUT', $transaction['request']->getMethod());
        $this->assertEquals('/api/user', $transaction['request']->getUri()->getPath());

        $requestBody = $transaction['request']->getBody()->getContents();
        $requestData = json_decode($requestBody);
        $this->assertEquals('user1', $requestData->n);
        $this->assertEquals('user123', $requestData->p);
    }

    /**
     * @dataProvider addAUserErrorProvider
     */
    public function testAddAUserError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->addAUser('test-token', 'user1', 'user123');
    }

    public function testAddAUserConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('PUT', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->addAUser('test-token', 'user1', 'user123');
    }

    public static function addAUserErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }

    public function testChangeUserPassword(): void
    {
        $this->mock->append(new Response(200));
        $this->apiClient->changeUserPassword('test-token', 1, 'new password');

        $this->assertCount(1, $this->container);

        $request = $this->container[0]['request'];
        $this->assertEquals('PATCH', $request->getMethod());
        $this->assertEquals('/api/user/1/password', $request->getUri()->getPath());

        $body = $request->getBody()->getContents();
        $data = json_decode($body);
        $this->assertTrue(isset($data->p));
        $this->assertEquals('new password', $data->p);
    }

    /**
     * @dataProvider changeUserPasswordErrorProvider
     */
    public function testChangeUserPasswordError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));
        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);
        $this->apiClient->changeUserPassword('test-token', 1, 'new password');
    }

    public static function changeUserPasswordErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }

    public function testChangeUserPasswordConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('PATCH', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->apiClient->changeUserPassword('test-token', 1, 'new password');
    }

    public function testDeleteSomeUsers(): void
    {
        $this->mock->append(new Response(200));

        $this->apiClient->deleteSomeUsers('test-token', [1, 2]);

        $this->assertCount(1, $this->container);

        $request = $this->container[0]['request'];
        $this->assertEquals('DELETE', $request->getMethod());
        $this->assertEquals('/api/users', $request->getUri()->getPath());

        $body = $request->getBody()->getContents();
        $data = json_decode($body);
        $this->assertTrue(isset($data->u));
        $this->assertCount(2, $data->u);
        $this->assertTrue(is_array($data->u));
        $this->assertEquals([1, 2], $data->u);
    }

    /**
     * @dataProvider deleteSomeUsersErrorProvider
     */
    public function testDeleteSomeUsersError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->deleteSomeUsers('test-token', [1, 2]);
    }

    public static function deleteSomeUsersErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }

    public function testDeleteSomeUsersConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('DELETE', '/test')));
        $this->expectException(TestcenterConnectorException::class);
        $this->apiClient->deleteSomeUsers('test-token', [1, 2]);
    }

    public function testDeleteSomeWorkspaces(): void
    {
        $this->mock->append(new Response(200));

        $this->apiClient->deleteSomeWorkspaces('test-token', [1, 2]);

        $this->assertEquals(1, count($this->container));

        $request = $this->container[0]['request'];
        $this->assertEquals('DELETE', $request->getMethod());
        $this->assertEquals('/api/workspaces', $request->getUri()->getPath());

        $body = $request->getBody()->getContents();
        $data = json_decode($body);
        $this->assertTrue(isset($data->ws));
        $this->assertCount(2, $data->ws);
        $this->assertTrue(is_array($data->ws));
        $this->assertEquals([1, 2], $data->ws);
    }

    /**
     * @dataProvider deleteSomeWorkspacesErrorProvider
     */
    public function testDeleteSomeWorkspacesError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->deleteSomeWorkspaces('test-token', [1, 2]);
    }

    public static function deleteSomeWorkspacesErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }

    public function testDeleteSomeWorkspacesConnectError(): void
    {
        $this->mock->append(new ConnectException('Connection error', new Request('DELETE', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->deleteSomeWorkspaces('test-token', [1, 2]);
    }

    public function testRenameAWorkspace(): void
    {
        $this->mock->append(new Response(200));

        $this->apiClient->renameAWorkspace('test-token', 1, 'renamed-workspace');

        $this->assertCount(1, $this->container);

        $transaction = $this->container[0];

        $this->assertEquals('PATCH', $transaction['request']->getMethod());
        $this->assertEquals('/api/workspace/1', $transaction['request']->getUri()->getPath());

        $requestBody = $transaction['request']->getBody()->getContents();
        $requestData = json_decode($requestBody);
        $this->assertEquals('renamed-workspace', $requestData->name);
    }

    /**
     * @dataProvider renameAWorkspaceErrorProvider
     */
    public function testRenameAWorkspaceError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->renameAWorkspace('test-token', 1, 'renamed-workspace');
    }

    public function testRenameAWorkspaceConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('PATCH', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->renameAWorkspace('test-token', 1, 'renamed-workspace');
    }

    public static function renameAWorkspaceErrorProvider(): array
    {
        return [
            [400], [401], [403], [410], [500]
        ];
    }

    public function testGetResults(): void
    {
        $responseBody = file_get_contents('testAssets/responses/get_results_valid.json');
        $this->mock->append(new Response(200, [], $responseBody));

        $results = $this->apiClient->getResults('test-token', 1);

        $this->assertCount(1, $results);

        $firstResult = $results[0];
        $this->assertEquals('sample_group', $firstResult->getGroupName());
        $this->assertEquals(2, $firstResult->getBookletsStarted());
        $this->assertEquals(0, $firstResult->getNumUnitsMin());
        $this->assertEquals(1, $firstResult->getNumUnitsMax());
        $this->assertEquals(1, $firstResult->getNumUnitsTotal());
        $this->assertEquals(0.5, $firstResult->getNumUnitsAvg());
        $this->assertEquals(new DateTime('2024-05-02T06:10:44'), $firstResult->getLastChange());
    }

    /**
     * @dataProvider getResultsMissingAttributeProvider
     */
    public function testGetResultsAttributeMissing(string $missingAttribute, string $responseBody): void
    {
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage($missingAttribute);

        $this->apiClient->getResults('test-token', 1);
    }

    public static function getResultsMissingAttributeProvider(): array
    {
        $attributes = ['groupName'];

        $validBody = file_get_contents('testAssets/responses/get_results_valid.json');
        $validBody = json_decode($validBody);

        $invalidBodies = [];
        foreach ($attributes as $attribute) {
            $firstResult = clone $validBody[0];
            unset($firstResult->{$attribute});
            $invalidBodies[] = [$attribute, json_encode([$firstResult])];
        }

        return $invalidBodies;
    }

    public function testGetResultsServerError(): void
    {
        $this->mock->append(new Response(500));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);

        $this->apiClient->getResults('test-token', 1);
    }

    public function testGetResultsConnectionError(): void
    {
        $this->mock->append(new ConnectException('Connection error', new Request('GET', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->getResults('test-token', 1);
    }

    public function testDeleteData(): void
    {
        $this->mock->append(new Response(200));

        $this->apiClient->deleteData('test-token', 1, ['test-group-1', 'test-group-2']);

        $this->assertCount(1, $this->container);

        $transaction = $this->container[0];

        $this->assertEquals('DELETE', $transaction['request']->getMethod());
        $this->assertEquals('/api/workspace/1/responses', $transaction['request']->getUri()->getPath());

        $requestBody = $transaction['request']->getBody()->getContents();
        $requestData = json_decode($requestBody);
        $this->assertCount(2, $requestData->groups);
        $this->assertContains('test-group-1', $requestData->groups);
        $this->assertContains('test-group-2', $requestData->groups);
    }

    /**
     * @dataProvider deleteDataErrorProvider
     */
    public function testDeleteDataError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->deleteData('test-token', 1, ['test-group']);
    }

    public static function deleteDataErrorProvider(): array
    {
        return [
            [401], [403], [404], [410], [500]
        ];
    }

    public function testDeleteDataConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('DELETE', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->deleteData('test-token', 1, ['test-group']);
    }

    public function testGetReportOfSystemChecks(): void
    {
        $responseBody = file_get_contents('testAssets/responses/report_system_checks_valid.json');
        $this->mock->append(new Response(200, [], $responseBody));

        $reports = $this->apiClient->getReportOfSystemChecks('test-token', 1, ['test-id']);
        $this->assertCount(1, $reports);

        $firstReport = $reports[0];
        $this->assertEquals('2020-02-17 13:01:31', $firstReport->getDate()->format('Y-m-d H:i:s'));
        $this->assertEquals('SYSCHECK.SAMPLE', $firstReport->getCheckId());
        $this->assertEquals('An example SysCheck definition', $firstReport->getCheckLabel());
        $this->assertCount(11, $firstReport->getEnvironment());
    }

    public function testGetReportOfSystemChecksInvalid(): void
    {
        $responseBody = file_get_contents('testAssets/responses/report_system_checks_invalid.json');
        $this->mock->append(new Response(200, [], $responseBody));

        $this->expectException(TestcenterConnectorException::class);
        $this->expectExceptionMessage('Field "checkId" is missing in response');

        $reports = $this->apiClient->getReportOfSystemChecks('test-token', 1, ['test-id']);
        $this->assertCount(1, $reports);
    }

    /**
     * @dataProvider getReportOfSystemChecksClientErrorProvider
     */
    public function testGetReportOfSystemChecksClientError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->getReportOfSystemChecks('test-token', 1, ['test']);
    }

    public function testGetReportOfSystemChecksServerError(): void
    {
        $this->mock->append(new Response(500));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode(500);

        $this->apiClient->getReportOfSystemChecks('test-token', 1, ['test']);
    }

    public function testGetReportOfSystemChecksConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('GET', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->getReportOfSystemChecks('test-token', 1, ['test']);
    }

    public function getReportOfSystemChecksClientErrorProvider(): array
    {
        return [
            [401], [403], [410]
        ];
    }

    public function testGetAListOfUsers(): void
    {
        $this->mock->append(new Response(200, [], '[
            {
                "name": "super",
                "id": 1,
                "email": null,
                "isSuperadmin": true
            }
        ]'));

        $users = $this->apiClient->getAListOfUsers('test-token');

        $this->assertCount(1, $users);
        $this->assertEquals('super', $users[0]->getName());
        $this->assertEquals(1, $users[0]->getId());
        $this->assertEquals(null, $users[0]->getEmail());
        $this->assertTrue($users[0]->isSuperadmin());
    }

    /**
     * @dataProvider getAListOfUsersErrorProvider
     */
    public function testGetAListOfUsersError(int $statusCode): void
    {
        $this->mock->append(new Response($statusCode));

        $this->expectException(ApiException::class);
        $this->expectExceptionCode($statusCode);

        $this->apiClient->getAListOfUsers('test-token');
    }

    public function testGetAListOfUsersConnectError(): void
    {
        $this->mock->append(new ConnectException('Connect error', new Request('GET', '/test')));

        $this->expectException(TestcenterConnectorException::class);

        $this->apiClient->getAListOfUsers('test-token');
    }

    public function testChangeSuperAdminStatusEnable(): void
    {
        $this->mock->append(new Response(200, [], ''));

        $success = $this->apiClient->changeSuperAdminStatus('test-token', 2, true, 'user123');

        $this->assertTrue($success);

        $transaction = $this->container[0];

        $this->assertEquals('PATCH', $transaction['request']->getMethod());
        $this->assertEquals('/api/user/2/super-admin/on', $transaction['request']->getUri()->getPath());
    }

    public function testChangeSuperAdminStatusDisable(): void
    {
        $this->mock->append(new Response(200, [], ''));

        $success = $this->apiClient->changeSuperAdminStatus('test-token', 2, false, 'user123');

        $this->assertTrue($success);

        $transaction = $this->container[0];

        $this->assertEquals('PATCH', $transaction['request']->getMethod());
        $this->assertEquals('/api/user/2/super-admin/off', $transaction['request']->getUri()->getPath());
    }

    public static function getAListOfUsersErrorProvider(): array
    {
        return [
            [401], [403], [410], [500]
        ];
    }
}
