<?php

namespace Kompetenztestde\TestcenterConnector;

use Generator;
use GuzzleHttp\Promise\FulfilledPromise;
use GuzzleHttp\Promise\Promise;
use GuzzleHttp\Psr7\Response;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileDeletionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\FileUploadResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\LogEntry;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SessionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\StartSessionResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\SystemConfigResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UnitResponse;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserRole;
use Kompetenztestde\TestcenterConnector\Api\Schemas\UserWorkspace;
use Kompetenztestde\TestcenterConnector\Api\Schemas\Workspace;
use Kompetenztestde\TestcenterConnector\Exceptions\ApiException;
use Kompetenztestde\TestcenterConnector\Exceptions\TestcenterConnectorException;
use PHPUnit\Framework\TestCase;
use Kompetenztestde\TestcenterConnector\Api\TestcenterApiClient;
use Kompetenztestde\TestcenterConnector\Models\File;
use Kompetenztestde\TestcenterConnector\Xml\XmlGenerator;
use Kompetenztestde\TestcenterConnector\Models\TestGroup;
use Kompetenztestde\TestcenterConnector\Models\Pupil;
use Kompetenztestde\TestcenterConnector\Models\Teacher;
use Kompetenztestde\TestcenterConnector\Models\User;
use Kompetenztestde\TestcenterConnector\Models\TestInterface;
use Kompetenztestde\TestcenterConnector\Xml\TestGroup as XmlTestGroup;
use Kompetenztestde\TestcenterConnector\Xml\TestLogin as XmlTestLogin;
use Kompetenztestde\TestcenterConnector\Xml\LoginMode as XmlLoginMode;

class TestcenterConnectorTest extends TestCase
{
    private $apiClient;
    private $xmlGenerator;

    /**
     * @val string
     */
    private $bookletId;

    /**
     * @val TestGroup
     */
    private $group;

    /**
     * @val string
     */
    private $authToken;

    /**
     * @val int
     */
    private $workspaceId;

    /**
     * @val string
     */
    private $testtakersContent;

    /**
     * @val Login[]
     */
    private $logins;

    /**
     * @var User
     */
    private $user;

    protected function setUp(): void
    {
        $this->apiClient = $this->createMock(TestcenterApiClient::class);
        $this->xmlGenerator = $this->createMock(XmlGenerator::class);
        $this->logins = [
            new Pupil('login1', 'password1'),
            new Teacher('login2', 'password2')
        ];
        $this->bookletId = 'booklet1';
        $this->group = new TestGroup('test-group', 'Test Group', $this->logins);
        $this->authToken = 'test-token';
        $this->workspaceId = 1;
        $this->testtakersContent = 'test content';
        $this->user = new User('super', 'user123', User::ADMIN_ROLE);
    }

    public function testAddTestGroup(): void
    {
        $this->xmlGenerator->expects($this->once())
            ->method('generateTesttakersXml')
            ->willReturn($this->testtakersContent);

        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('uploadFileAsync')
            ->with(
                $this->identicalTo($this->authToken),
                $this->identicalTo($this->workspaceId),
                $this->identicalTo('testtakers-test-group.xml'),
                $this->testtakersContent
            )
            ->willReturn(new FulfilledPromise(new FileUploadResponse([])));

        $configResponse = new SystemConfigResponse(
            (object) [
                'version' => '15.5.0'
            ]
        );

        $this->apiClient->expects($this->once())
            ->method('getSystemConfigAsync')
            ->willReturn(new FulfilledPromise($configResponse));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->addTestGroup($this->workspaceId, $this->group, $this->bookletId, $this->xmlGenerator);
    }

    public function testAddTestGroupGroupPassword(): void
    {
        $logins = [
            new Pupil('login1')
        ];
        $group = new TestGroup('test-group', 'Test Group', $logins, 'grouppassword');
        $expectedGroup = new XmlTestGroup(
            $this->group->getId(),
            $this->group->getLabel(),
            [
                new XmlTestLogin('grouppassword-login1', null, XmlLoginMode::RUN_HOT_RETURN)
            ],
            ['test-booklet'],
            null,
            null
        );

        $this->xmlGenerator->expects($this->once())
            ->method('generateTesttakersXml')
            ->with($this->equalTo($expectedGroup))
            ->willReturn($this->testtakersContent);

        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('uploadFileAsync')
            ->willReturn(new FulfilledPromise(new FileUploadResponse([])));

        $configResponse = new SystemConfigResponse(
            (object) [
                'version' => '15.5.0'
            ]
        );

        $this->apiClient->expects($this->once())
            ->method('getSystemConfigAsync')
            ->willReturn(new FulfilledPromise($configResponse));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->addTestGroup($this->workspaceId, $group, 'test-booklet', $this->xmlGenerator);
    }

    /**
     * Test that multiple addTestGroup calls only get one auth token
     */
    public function testAddTestGroupReuseAuthToken(): void
    {
        $this->xmlGenerator->expects($this->exactly(2))
            ->method('generateTesttakersXml')
            ->willReturn($this->testtakersContent);

        $this->successfulLogin();

        $this->apiClient->expects($this->exactly(2))
            ->method('uploadFileAsync')
            ->with(
                $this->identicalTo($this->authToken),
                $this->identicalTo($this->workspaceId),
                $this->identicalTo('testtakers-test-group.xml'),
                $this->testtakersContent
            )
            ->willReturn(new FulfilledPromise(new FileUploadResponse([])));

        $sessionResponse = new SessionResponse(
            (object) [
                'token' => $this->authToken,
                'displayName' => 'Test User'
            ]
        );

        $this->apiClient->expects($this->once())
            ->method('getASessionAsync')
            ->with($this->identicalTo($this->authToken))
            ->willReturn(new FulfilledPromise($sessionResponse));

        $configResponse = new SystemConfigResponse(
            (object) [
                'version' => '15.5.0'
            ]
        );

        $this->apiClient->expects($this->any())
            ->method('getSystemConfigAsync')
            ->willReturn(new FulfilledPromise($configResponse));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->addTestGroup($this->workspaceId, $this->group, $this->bookletId, $this->xmlGenerator);
        $connector->addTestGroup($this->workspaceId, $this->group, $this->bookletId, $this->xmlGenerator);
    }

    /**
     * Test that multiple addTestGroup calls only request a new auth token if it
     * has expired
     */
    public function testAddTestGroupExpiredAuthToken(): void
    {
        $this->xmlGenerator->expects($this->exactly(2))
            ->method('generateTesttakersXml')
            ->willReturn($this->testtakersContent);

        $this->successfulLogin();

        $this->apiClient->expects($this->exactly(2))
            ->method('uploadFileAsync')
            ->with(
                $this->identicalTo($this->authToken),
                $this->identicalTo($this->workspaceId),
                $this->identicalTo('testtakers-test-group.xml'),
                $this->testtakersContent
            )
            ->willReturn(new FulfilledPromise(new FileUploadResponse([])));

        $sessionResponse = new Promise(function () {
            throw new ApiException("Expired", 409);
        });

        $this->apiClient->expects($this->once())
            ->method('getASessionAsync')
            ->with(
                $this->identicalTo($this->authToken)
            )
            ->willReturn($sessionResponse);

        $configResponse = new SystemConfigResponse(
            (object) [
                'version' => '15.5.0'
            ]
        );

        $this->apiClient->expects($this->any())
            ->method('getSystemConfigAsync')
            ->willReturn(new FulfilledPromise($configResponse));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->addTestGroup($this->workspaceId, $this->group, $this->bookletId, $this->xmlGenerator);
        $connector->addTestGroup($this->workspaceId, $this->group, $this->bookletId, $this->xmlGenerator);
    }

    public function testAddWorkspace(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('addAWorkspaceAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo('test-workspace')
            )
            ->willReturn(new FulfilledPromise(2));

        $this->apiClient->expects($this->never())->method('changeUserRoles');

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $workspaceId = $connector->addWorkspace('test-workspace');

        $this->assertEquals(2, $workspaceId);
    }

    public function testAddWorkspaceSetRole(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('addAWorkspaceAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo('test-workspace')
            )
            ->willReturn(new FulfilledPromise(2));

        $users = [new \Kompetenztestde\TestcenterConnector\Api\Schemas\User((object) [
            'name' => 'super',
            'id' => 1
        ])];

        $this->apiClient->expects($this->once())
            ->method('getAListOfUsersAsync')
            ->with($this->authToken)
            ->willReturn(new FulfilledPromise($users));

        $this->apiClient->expects($this->once())
            ->method('changeUserRolesAsync')
            ->with(
                $this->authToken,
                2,
                [new UserRole(1, UserRole::RW)]
            )
            ->willReturn(new FulfilledPromise(true));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $workspaceId = $connector->addWorkspace('test-workspace', true);

        $this->assertEquals(2, $workspaceId);
    }

    public function testAddWorkspaceSetRoleReturnUserId(): void
    {
        $startSessionResponse = new StartSessionResponse(
            (object) [
                'token' => $this->authToken,
                'displayName' => 'super',
                'id' => 1
            ]
        );
        $this->apiClient->expects($this->once())
            ->method('startAdminSessionAsync')
            ->willReturn(new FulfilledPromise($startSessionResponse));

        $this->apiClient->expects($this->once())
            ->method('addAWorkspaceAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo('test-workspace')
            )
            ->willReturn(new FulfilledPromise(2));

        $this->apiClient->expects($this->once())
            ->method('changeUserRolesAsync')
            ->with(
                $this->authToken,
                2,
                [new UserRole(1, UserRole::RW)]
            )
            ->willReturn(new FulfilledPromise(true));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $workspaceId = $connector->addWorkspace('test-workspace', true);

        $this->assertEquals(2, $workspaceId);
    }

    public function testAddWorkspaceLegacy(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('addAWorkspaceAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo('test-workspace')
            )
            ->willReturn(new FulfilledPromise(null));

        $workspaces = [
            new Workspace(
                (object) [
                    'id' => 1,
                    'name' => 'test-workspace'
                ]
            ),
            new Workspace(
                (object) [
                    'id' => 2,
                    'name' => 'test-workspace-2'
                ]
            )
        ];

        $this->apiClient->expects($this->once())
            ->method('getAListOfWorkspacesAsync')
            ->with($this->equalTo($this->authToken))
            ->willReturn(new FulfilledPromise($workspaces));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $workspaceId = $connector->addWorkspace('test-workspace');

        $this->assertEquals(1, $workspaceId);
    }

    public function testGetWorkspaces(): void
    {
        $this->successfulLogin();

        $workspaces = [
            new Workspace((object) [
                'id' => 1,
                'name' => 'workspace-1'
            ]),
            new Workspace((object) [
                'id' => 2,
                'name' => 'workspace-2'
            ])
        ];

        $this->apiClient->expects($this->once())
            ->method('getAListOfWorkspacesAsync')
            ->willReturn(new FulfilledPromise($workspaces));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $workspaces = $connector->getWorkspaces();

        $this->assertCount(2, $workspaces);
        $this->assertEquals(1, $workspaces[0]->getId());
        $this->assertEquals('workspace-1', $workspaces[0]->getName());
        $this->assertEquals(2, $workspaces[1]->getId());
        $this->assertEquals('workspace-2', $workspaces[1]->getName());
    }

    public function testDeleteWorkspace(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('deleteSomeWorkspacesAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo([1])
            )
            ->willReturn(new FulfilledPromise(true));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->deleteWorkspace([1]);
    }

    public function testDeleteWorkspaceError(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('deleteSomeWorkspacesAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo([1])
            )
            ->willThrowException(new ApiException('API Error', 400, [], 'test body'));

        $this->expectException(TestcenterConnectorException::class);

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->deleteWorkspace([1]);
    }

    public function testAddTest(): void
    {
        $test = $this->createStub(TestInterface::class);
        $test->method('getBookletFilename')
            ->willReturn('booklet.xml');

        $test->method('getBookletContent')
            ->willReturn('test booklet');

        $test->method('getUnits')
            ->willReturn($this->arrayAsGenerator([
                'unit1.xml' => 'test unit'
            ]));

        $test->method('getResources')
            ->willReturn($this->arrayAsGenerator([
                'resource1.html' => 'test resource'
            ]));


        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('uploadFile')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo('test.zip'),
                $this->anything()
            );

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $connector->addTest(1, $test);
    }

    public function testGetGroupResults(): void
    {
        $testResponses = [
            new UnitResponse(
                (object) [
                    'groupname' => 'test-group',
                    'loginname' => 'test-user',
                    'code' => '',
                    'bookletname' => 'test-booklet',
                    'unitname' => 'test-unit',
                    'laststate' => '',
                    'responses' => [
                        (object) [
                            'id' => 'test',
                            'content' => 'test1',
                            'ts' => 123,
                            'responseType' => 'test'
                        ]
                    ]
                ]
            )
        ];

        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('getReportOfItemResponsesAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo(['test-group'])
            )
            ->willReturn(new FulfilledPromise($testResponses));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $results = $connector->getGroupResults(1, 'test-group');

        $this->assertEquals($testResponses, $results);
    }

    public function testGetWorkspacesByUser(): void
    {
        $this->successfulLogin();

        $workspaces = [
            new UserWorkspace(
                (object) [
                    'id' => 1,
                    'name' => 'Workspace 1',
                    'role' => 'RW',
                    'selected' => true
                ]
            ),
            new UserWorkspace(
                (object) [
                    'id' => 2,
                    'name' => 'Workspace 2',
                    'role' => 'RO',
                    'selected' => false
                ]
            )
        ];
        $this->apiClient->expects($this->once())
            ->method('getAListOfUserWorkspaces')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1)
            )
            ->willReturn($workspaces);

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $returnedWorkspaces = $connector->getWorkspacesByUser(1);

        $this->assertEquals($workspaces, $returnedWorkspaces);
    }

    public function testChangeUserRoles(): void
    {
        $this->successfulLogin();

        $userRoles = [
            new UserRole(
                1,
                UserRole::RW
            ),
            new UserRole(
                2,
                UserRole::RO
            )
        ];
        $this->apiClient->expects($this->once())
            ->method('changeUserRoles')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo($userRoles)
            )
            ->willReturn(true);

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $response = $connector->changeUserRoles(1, $userRoles);

        $this->assertTrue($response);
    }

    public function testGetLogs(): void
    {
        $this->successfulLogin();

        $logs = [
            new LogEntry(
                (object) [
                    'groupname' => 'test_group',
                    'loginname' => 'test_login',
                    'code' => 'abc',
                    'bookletname' => 'test_booklet',
                    'unitname' => 'test_unit',
                    'timestamp' => 123,
                    'logentry' => 'Test entry'
                ]
            )
        ];
        $this->apiClient->expects($this->once())
            ->method('getReportOfLogsAsync')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo(['test_group'])
            )
            ->willReturn(new FulfilledPromise($logs));

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $returnedLogs = $connector->getLogs(1, ['test_group']);

        $this->assertEquals($logs, $returnedLogs);
    }

    public function testDeleteFiles(): void
    {
        $this->successfulLogin();

        $deletionResponse = new FileDeletionResponse(
            (object) [
                'deleted' => [
                    'Testtakers/SAMPLE_TESTTAKERS.XML'
                ],
                'did_not_exist' => [],
                'not_allowed' => [],
                'was_used' => []
            ]
        );
        $this->apiClient->expects($this->once())
            ->method('deleteFiles')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo([
                    'Testtakers/SAMPLE_TESTTAKERS.XML',
                    'Unit/UNIT1.XML'
                ])
            )
            ->willReturn($deletionResponse);

        $files = [
            new File('SAMPLE_TESTTAKERS.XML', File::TESTTAKERS_FILE),
            new File('UNIT1.XML', File::UNIT_FILE)
        ];

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $response = $connector->deleteFiles(1, $files);

        $this->assertEquals($deletionResponse, $response);
    }

    public function testChangeUserPassword(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('changeUserPassword')
            ->with(
                $this->equalTo($this->authToken),
                $this->equalTo(1),
                $this->equalTo('new-password')
            )
            ->willReturn(true);

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $result = $connector->changeUserPassword(1, 'new-password');

        $this->assertTrue($result);
    }

    public function testAddSuperAdmin(): void
    {
        $this->successfulLogin();

        $this->apiClient->expects($this->once())
            ->method('addAUser')
            ->with(
                $this->authToken,
                'super2',
                'user1234'
            )
            ->willReturn(true);

        $this->apiClient->expects($this->once())
            ->method('getAListOfUsers')
            ->with($this->authToken)
            ->willReturn([
                new \Kompetenztestde\TestcenterConnector\Api\Schemas\User((object) [
                    'name' => 'super1',
                    'id' => 1
                ]),
                new \Kompetenztestde\TestcenterConnector\Api\Schemas\User((object) [
                    'name' => 'super2',
                    'id' => 2
                ])
            ]);

        $this->apiClient->expects($this->once())
            ->method('changeSuperAdminStatus')
            ->with(
                $this->authToken,
                2,
                true,
                $this->user->getPassword()
            )
            ->willReturn(true);

        $this->apiClient->expects($this->once())
            ->method('getAListOfWorkspaces')
            ->with($this->authToken)
            ->willReturn([
                new Workspace((object) [
                    'id' => 1,
                    'name' => 'workspace-1'
                ]),
                new Workspace((object) [
                    'id' => 2,
                    'name' => 'workspace-2'
                ])
            ]);

        $this->apiClient->expects($this->exactly(2))
            ->method('changeUserRoles')
            ->withConsecutive(
                [
                    $this->authToken, 1, [new UserRole(2, UserRole::RW)]
                ],
                [
                    $this->authToken, 2, [new UserRole(2, UserRole::RW)]
                ]
            )
            ->willReturn(true);

        $connector = new TestcenterConnector($this->user, null, $this->apiClient);
        $result = $connector->addSuperAdmin('super2', 'user1234');

        $this->assertEquals(2, $result);
    }

    private function arrayAsGenerator(array $array): Generator
    {
        foreach ($array as $key => $item) {
            yield $key => $item;
        }
    }

    private function successfulLogin()
    {
        $startSessionResponse = new StartSessionResponse((object) [
            'token' => $this->authToken,
            'displayName' => $this->user->getUsername()
        ]);
        $this->apiClient->expects($this->atLeastOnce())
            ->method('startAdminSessionAsync')
            ->willReturn(new FulfilledPromise($startSessionResponse));
    }
}
